"""Contains API representation for the MySkoda REST API."""

import asyncio
import json
import logging
from collections.abc import Callable
from dataclasses import dataclass
from datetime import UTC, datetime
from urllib.parse import quote

from aiohttp import ClientResponseError, ClientSession

from myskoda.anonymize import (
    anonymize_air_conditioning,
    anonymize_auxiliary_heating,
    anonymize_charging,
    anonymize_chargingprofiles,
    anonymize_departure_timers,
    anonymize_driving_range,
    anonymize_garage,
    anonymize_health,
    anonymize_info,
    anonymize_maintenance,
    anonymize_positions,
    anonymize_status,
    anonymize_trip_statistics,
    anonymize_url,
    anonymize_user,
    anonymize_vehicle_connection_status,
)
from myskoda.models.charging import ChargeMode
from myskoda.models.garage import Garage
from myskoda.models.position import Position, PositionType

from .auth.authorization import Authorization
from .const import BASE_URL_SKODA, REQUEST_TIMEOUT_IN_SECONDS
from .models.air_conditioning import (
    AirConditioning,
    AirConditioningAtUnlock,
    AirConditioningTimer,
    AirConditioningWithoutExternalPower,
    SeatHeating,
    WindowHeating,
)
from .models.auxiliary_heating import AuxiliaryConfig, AuxiliaryHeating, AuxiliaryHeatingTimer
from .models.charging import Charging
from .models.chargingprofiles import ChargingProfiles
from .models.departure import DepartureInfo, DepartureTimer
from .models.driving_range import DrivingRange
from .models.health import Health
from .models.info import Info
from .models.maintenance import Maintenance, MaintenanceReport
from .models.position import Positions
from .models.spin import Spin
from .models.status import Status
from .models.trip_statistics import TripStatistics
from .models.user import User
from .models.vehicle_connection_status import VehicleConnectionStatus

_LOGGER = logging.getLogger(__name__)


@dataclass
class GetEndpointResult[T]:
    url: str
    raw: str
    result: T


class RestApi:
    """API hub class that can perform all calls to the MySkoda API."""

    session: ClientSession
    authorization: Authorization

    def __init__(self, session: ClientSession, authorization: Authorization) -> None:
        self.session = session
        self.authorization = authorization

    def process_json(
        self,
        data: str,
        anonymize: bool,
        anonymization_fn: Callable[[dict], dict],
    ) -> str:
        """Process the raw json returned by the API with some preprocessor logic."""
        if not anonymize:
            return data
        parsed = json.loads(data)
        anonymized = anonymization_fn(parsed)
        return json.dumps(anonymized)

    async def _make_request(self, url: str, method: str, json: dict | None = None) -> str:
        try:
            async with asyncio.timeout(REQUEST_TIMEOUT_IN_SECONDS):
                async with self.session.request(
                    method=method,
                    url=f"{BASE_URL_SKODA}/api{url}",
                    headers=await self._headers(),
                    json=json,
                ) as response:
                    await response.text()  # Ensure response is fully read
                    response.raise_for_status()
                    return await response.text()
        except TimeoutError:
            _LOGGER.exception("Timeout while sending %s request to %s", method, url)
            raise
        except ClientResponseError as err:
            _LOGGER.exception("Invalid status for %s request to %s: %d", method, url, err.status)
            raise

    async def _make_get_request[T](self, url: str) -> str:
        return await self._make_request(url=url, method="GET")

    async def _make_post_request(self, url: str, json: dict | None = None) -> str:
        return await self._make_request(url=url, method="POST", json=json)

    async def _make_put_request(self, url: str, json: dict | None = None) -> str:
        return await self._make_request(url=url, method="PUT", json=json)

    async def verify_spin(self, spin: str, anonymize: bool = False) -> GetEndpointResult[Spin]:
        """Verify SPIN."""
        url = "/v1/spin/verify"
        json_data = {"currentSpin": spin}
        raw = self.process_json(
            data=await self._make_post_request(url, json_data),
            anonymize=anonymize,
            anonymization_fn=anonymize_info,
        )
        result = self._deserialize(raw, Spin.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_info(self, vin: str, anonymize: bool = False) -> GetEndpointResult[Info]:
        """Retrieve information related to basic information for the specified vehicle."""
        url = f"/v2/garage/vehicles/{vin}?connectivityGenerations=MOD1&connectivityGenerations=MOD2&connectivityGenerations=MOD3&connectivityGenerations=MOD4"  # noqa: E501
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_info,
        )
        result = self._deserialize(raw, Info.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_charging(self, vin: str, anonymize: bool = False) -> GetEndpointResult[Charging]:
        """Retrieve information related to charging for the specified vehicle."""
        url = f"/v1/charging/{vin}"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_charging,
        )
        result = self._deserialize(raw, Charging.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_charging_profiles(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[ChargingProfiles]:
        """Retrieve information related to chargingprofiles for the specified vehicle."""
        url = f"/v1/charging/{vin}/profiles"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_chargingprofiles,
        )
        result = self._deserialize(raw, ChargingProfiles.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_status(self, vin: str, anonymize: bool = False) -> GetEndpointResult[Status]:
        """Retrieve the current status for the specified vehicle."""
        url = f"/v2/vehicle-status/{vin}"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_status,
        )
        result = self._deserialize(raw, Status.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_air_conditioning(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[AirConditioning]:
        """Retrieve the current air conditioning status for the specified vehicle."""
        url = f"/v2/air-conditioning/{vin}"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_air_conditioning,
        )
        result = self._deserialize(raw, AirConditioning.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_auxiliary_heating(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[AuxiliaryHeating]:
        """Retrieve the current auxiliary heating status for the specified vehicle."""
        url = f"/v2/air-conditioning/{vin}/auxiliary-heating"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_auxiliary_heating,
        )
        result = self._deserialize(raw, AuxiliaryHeating.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_positions(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[Positions]:
        """Retrieve the current position for the specified vehicle."""
        url = f"/v1/maps/positions?vin={vin}"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_positions,
        )
        result = self._deserialize(raw, Positions.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_driving_range(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[DrivingRange]:
        """Retrieve estimated driving range for combustion vehicles."""
        url = f"/v2/vehicle-status/{vin}/driving-range"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_driving_range,
        )
        result = self._deserialize(raw, DrivingRange.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_trip_statistics(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[TripStatistics]:
        """Retrieve statistics about past trips."""
        url = f"/v1/trip-statistics/{vin}?offsetType=week&offset=0&timezone=Europe%2FBerlin"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_trip_statistics,
        )
        result = self._deserialize(raw, TripStatistics.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_maintenance(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[Maintenance]:
        """Retrieve maintenance report, settings and history."""
        url = f"/v3/vehicle-maintenance/vehicles/{vin}"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_maintenance,
        )
        result = self._deserialize(raw, Maintenance.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_maintenance_report(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[MaintenanceReport]:
        """Retrieve just the maintenance report."""
        url = f"/v3/vehicle-maintenance/vehicles/{vin}/report"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_maintenance,
        )
        result = self._deserialize(raw, MaintenanceReport.from_json)
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_health(self, vin: str, anonymize: bool = False) -> GetEndpointResult[Health]:
        """Retrieve health information for the specified vehicle."""
        url = f"/v1/vehicle-health-report/warning-lights/{vin}"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_health,
        )
        result = self._deserialize(raw, Health.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_user(self, anonymize: bool = False) -> GetEndpointResult[User]:
        """Retrieve user information about logged in user."""
        url = "/v1/users"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_user,
        )
        result = self._deserialize(raw, User.from_json)
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_garage(self, anonymize: bool = False) -> GetEndpointResult[Garage]:
        """Fetch the garage (list of vehicles with limited info)."""
        url = "/v2/garage?connectivityGenerations=MOD1&connectivityGenerations=MOD2&connectivityGenerations=MOD3&connectivityGenerations=MOD4"  # noqa: E501
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_garage,
        )
        result = self._deserialize(raw, Garage.from_json)
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_departure_timers(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[DepartureInfo]:
        """Retrieve departure timers for the vehicle."""
        # Get the current local time with timezone
        now = datetime.now().astimezone()
        # Format the datetime string with timezone
        formatted_time = (
            now.strftime("%Y-%m-%dT%H:%M:%S.%f")
            + now.strftime("%z")[:3]
            + ":"
            + now.strftime("%z")[3:]
        )

        url = (
            f"/v1/vehicle-automatization/{vin}/departure/timers"
            f"?deviceDateTime={quote(formatted_time, safe='')}"
        )
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_departure_timers,
        )
        result = self._deserialize(raw, DepartureInfo.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def get_vehicle_connection_status(
        self, vin: str, anonymize: bool = False
    ) -> GetEndpointResult[VehicleConnectionStatus]:
        """Retrieve vehicle connection status."""
        url = f"/v2/connection-status/{vin}/readiness"
        raw = self.process_json(
            data=await self._make_get_request(url),
            anonymize=anonymize,
            anonymization_fn=anonymize_vehicle_connection_status,
        )
        result = self._deserialize(raw, VehicleConnectionStatus.from_json)
        url = anonymize_url(url) if anonymize else url
        return GetEndpointResult(url=url, raw=raw, result=result)

    async def _headers(self) -> dict[str, str]:
        return {"authorization": f"Bearer {await self.authorization.get_access_token()}"}

    async def stop_air_conditioning(self, vin: str) -> None:
        """Stop the air conditioning."""
        _LOGGER.debug("Stopping air conditioning for vehicle %s", vin)
        await self._make_post_request(url=f"/v2/air-conditioning/{vin}/stop")

    async def start_air_conditioning(self, vin: str, temperature: float) -> None:
        """Start the air conditioning."""
        round_temp = round(temperature * 2) / 2
        _LOGGER.debug(
            "Starting air conditioning for vehicle %s with temperature %.1f",
            vin,
            round_temp,
        )
        json_data = {
            "heaterSource": "ELECTRIC",
            "targetTemperature": {
                "temperatureValue": round_temp,
                "unitInCar": "CELSIUS",
            },
        }
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/start",
            json=json_data,
        )

    async def stop_ventilation(self, vin: str) -> None:
        """Start the ventilation."""
        _LOGGER.debug("Stopping ventilation for vehicle %s", vin)
        await self._make_post_request(url=f"/v2/air-conditioning/{vin}/active-ventilation/stop")

    async def start_ventilation(self, vin: str) -> None:
        """Start the ventilation."""
        _LOGGER.debug("Starting ventilation for vehicle %s", vin)
        await self._make_post_request(url=f"/v2/air-conditioning/{vin}/active-ventilation/start")

    async def stop_auxiliary_heating(self, vin: str) -> None:
        """Stop the auxiliary heating."""
        _LOGGER.debug("Stopping auxiliary heating for vehicle %s", vin)
        await self._make_post_request(url=f"/v2/air-conditioning/{vin}/auxiliary-heating/stop")

    async def start_auxiliary_heating(
        self, vin: str, spin: str, config: AuxiliaryConfig | None = None
    ) -> None:
        """Start the auxiliary heating."""
        _LOGGER.debug("Starting auxiliary heating for vehicle %s", vin)

        json_data: dict[str, object] = {"spin": spin}
        if config is not None:
            json_data = json_data | config.to_dict(omit_none=True, by_alias=True)

        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/auxiliary-heating/start",
            json=json_data,
        )

    async def set_ac_without_external_power(
        self, vin: str, settings: AirConditioningWithoutExternalPower
    ) -> None:
        """Enable or disable AC without external power."""
        _LOGGER.debug(
            "Setting AC without external power for vehicle %s to %r",
            vin,
            settings.air_conditioning_without_external_power_enabled,
        )
        json_data = settings.to_dict()
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/settings/ac-without-external-power",
            json=json_data,
        )

    async def set_ac_at_unlock(self, vin: str, settings: AirConditioningAtUnlock) -> None:
        """Enable or disable AC at unlock."""
        _LOGGER.debug(
            "Setting AC at at unlock for vehicle %s to %r",
            vin,
            settings.air_conditioning_at_unlock_enabled,
        )
        json_data = settings.to_dict()
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/settings/ac-at-unlock",
            json=json_data,
        )

    async def set_windows_heating(self, vin: str, settings: WindowHeating) -> None:
        """Enable or disable windows heating with AC."""
        _LOGGER.debug(
            "Setting windows heating with AC for vehicle %s to %r",
            vin,
            settings.window_heating_enabled,
        )
        json_data = settings.to_dict()
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/settings/windows-heating",
            json=json_data,
        )

    async def set_seats_heating(self, vin: str, settings: SeatHeating) -> None:
        """Enable or disable seats heating with AC."""
        json_data = settings.to_dict(omit_none=True, by_alias=True)
        _LOGGER.debug("Setting seats heating with AC for vehicle %s: %s", vin, json_data)
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/settings/seats-heating",
            json=json_data,
        )

    async def set_target_temperature(self, vin: str, temperature: float) -> None:
        """Set the air conditioning's target temperature in °C."""
        round_temp = round(temperature * 2) / 2
        _LOGGER.debug("Setting target temperature for vehicle %s to %.1f", vin, round_temp)
        json_data = {"temperatureValue": round_temp, "unitInCar": "CELSIUS"}
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/settings/target-temperature",
            json=json_data,
        )

    async def start_window_heating(self, vin: str) -> None:
        """Start heating both the front and rear window."""
        _LOGGER.debug("Starting window heating for vehicle %s", vin)
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/start-window-heating",
        )

    async def stop_window_heating(self, vin: str) -> None:
        """Stop heating both the front and rear window."""
        _LOGGER.debug("Stopping window heating for vehicle %s", vin)
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/stop-window-heating",
        )

    async def set_charge_limit(self, vin: str, limit: int) -> None:
        """Set the maximum charge limit in percent."""
        _LOGGER.debug("Setting charge limit for vehicle %s to %d", vin, limit)
        json_data = {"targetSOCInPercent": limit}
        await self._make_put_request(
            url=f"/v1/charging/{vin}/set-charge-limit",
            json=json_data,
        )

    async def set_minimum_charge_limit(self, vin: str, limit: int) -> None:
        """Set minimum battery SoC in percent for departure timer."""
        _LOGGER.debug(
            "Setting minimum SoC for departure timers for vehicle %s to %r",
            vin,
            limit,
        )

        json_data = {"minimumBatteryStateOfChargeInPercent": limit}
        await self._make_post_request(
            url=f"/v1/vehicle-automatization/{vin}/departure/timers/settings",
            json=json_data,
        )

    # TODO @dvx76: Maybe refactor for FBT001
    async def set_battery_care_mode(self, vin: str, enabled: bool) -> None:
        """Enable or disable the battery care mode."""
        _LOGGER.debug("Setting battery care mode for vehicle %s to %r", vin, enabled)
        json_data = {"chargingCareMode": "ACTIVATED" if enabled else "DEACTIVATED"}
        await self._make_put_request(
            url=f"/v1/charging/{vin}/set-care-mode",
            json=json_data,
        )

    # TODO @dvx76: Maybe refactor for FBT001
    async def set_auto_unlock_plug(self, vin: str, enabled: bool) -> None:
        """Enable or disable auto unlock plug when charged."""
        _LOGGER.debug("Setting auto unlock plug for vehicle %s to %r", vin, enabled)
        json_data = {"autoUnlockPlug": "PERMANENT" if enabled else "OFF"}
        await self._make_put_request(
            url=f"/v1/charging/{vin}/set-auto-unlock-plug",
            json=json_data,
        )

    # TODO @dvx76: Maybe refactor for FBT001
    async def set_reduced_current_limit(self, vin: str, reduced: bool) -> None:
        """Enable reducing the current limit by which the car is charged."""
        _LOGGER.debug("Setting reduced charging for vehicle %s to %r", vin, reduced)
        json_data = {"chargingCurrent": "REDUCED" if reduced else "MAXIMUM"}
        await self._make_put_request(
            url=f"/v1/charging/{vin}/set-charging-current",
            json=json_data,
        )

    async def start_charging(self, vin: str) -> None:
        """Start charging the car."""
        _LOGGER.debug("Starting charging for vehicle %s", vin)
        await self._make_post_request(
            url=f"/v1/charging/{vin}/start",
        )

    async def stop_charging(self, vin: str) -> None:
        """Stop charging the car."""
        _LOGGER.debug("Stopping charging of vehicle %s", vin)
        await self._make_post_request(
            url=f"/v1/charging/{vin}/stop",
        )

    async def wakeup(self, vin: str) -> None:
        """Wake the vehicle up. Can be called maximum three times a day."""
        _LOGGER.debug("Waking up vehicle %s", vin)
        await self._make_post_request(
            url=f"/v1/vehicle-wakeup/{vin}?applyRequestLimiter=true",
        )

    async def set_charge_mode(self, vin: str, mode: ChargeMode) -> None:
        """Wake the vehicle up. Can be called maximum three times a day."""
        _LOGGER.debug("Changing charging mode of vehicle %s to %s", vin, mode)
        json_data = {"chargeMode": mode.value}
        await self._make_post_request(
            url=f"/v1/charging/{vin}/set-charge-mode",
            json=json_data,
        )

    async def lock(self, vin: str, spin: str) -> None:
        """Lock the vehicle."""
        _LOGGER.debug("Locking vehicle %s", vin)
        json_data = {"currentSpin": spin}
        await self._make_post_request(
            url=f"/v1/vehicle-access/{vin}/lock",
            json=json_data,
        )

    async def unlock(self, vin: str, spin: str) -> None:
        """Unlock the vehicle."""
        _LOGGER.debug("Unlocking vehicle %s", vin)
        json_data = {"currentSpin": spin}
        await self._make_post_request(
            url=f"/v1/vehicle-access/{vin}/unlock",
            json=json_data,
        )

    # TODO @dvx76: Maybe refactor for FBT001
    async def honk_flash(
        self,
        vin: str,
        positions: list[Position],
    ) -> None:
        """Emit Honk and flash."""
        position = next(pos for pos in positions if pos.type == PositionType.VEHICLE)
        # TODO @webspider: Make this a proper class
        json_data = {
            "mode": "HONK_AND_FLASH",
            "vehiclePosition": {
                "latitude": position.gps_coordinates.latitude,
                "longitude": position.gps_coordinates.longitude,
            },
        }
        await self._make_post_request(
            url=f"/v1/vehicle-access/{vin}/honk-and-flash", json=json_data
        )

    async def flash(
        self,
        vin: str,
        positions: list[Position],
    ) -> None:
        """Emit flash."""
        position = next(pos for pos in positions if pos.type == PositionType.VEHICLE)
        # TODO @webspider: Make this a proper class
        json_data = {
            "mode": "FLASH",
            "vehiclePosition": {
                "latitude": position.gps_coordinates.latitude,
                "longitude": position.gps_coordinates.longitude,
            },
        }
        await self._make_post_request(
            url=f"/v1/vehicle-access/{vin}/honk-and-flash", json=json_data
        )

    async def set_departure_timer(self, vin: str, timer: DepartureTimer) -> None:
        """Set departure timer."""
        _LOGGER.debug(
            "Setting departure timer #%i for vehicle %s to %r", timer.id, vin, timer.enabled
        )

        now = datetime.now(UTC)
        datetime_str = now.isoformat()

        json_data = {"deviceDateTime": datetime_str, "timers": [timer.to_dict()]}
        await self._make_post_request(
            url=f"/v1/vehicle-automatization/{vin}/departure/timers",
            json=json_data,
        )

    async def set_ac_timer(self, vin: str, timer: AirConditioningTimer) -> None:
        """Set air-conditioning timer."""
        _LOGGER.debug(
            "Setting air-conditioning timer #%i for vehicle %s to %r", timer.id, vin, timer.enabled
        )

        json_data = {"timers": [timer.to_dict(by_alias=True)]}
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/timers",
            json=json_data,
        )

    async def set_auxiliary_heating_timer(
        self, vin: str, timer: AuxiliaryHeatingTimer, spin: str
    ) -> None:
        """Set auxiliary heating timer."""
        _LOGGER.debug(
            "Setting auxiliary heating timer #%i for vehicle %s to %r", timer.id, vin, timer.enabled
        )

        json_data = {"spin": spin, "timers": [timer.to_dict(by_alias=True)]}
        await self._make_post_request(
            url=f"/v2/air-conditioning/{vin}/auxiliary-heating/timers",
            json=json_data,
        )

    def _deserialize[T](self, text: str, deserialize: Callable[[str], T]) -> T:
        try:
            data = deserialize(text)
        except Exception:
            _LOGGER.exception("Failed to deserialize data: %s", text)
            raise
        else:
            return data



"""MySkoda Vehicle connection status."""

from dataclasses import dataclass, field

from mashumaro import field_options

from .common import BaseResponse


@dataclass
class VehicleConnectionStatus(BaseResponse):
    unreachable: bool
    in_motion: bool = field(metadata=field_options(alias="inMotion"))
    battery_protection_limit_on: bool = field(
        metadata=field_options(alias="batteryProtectionLimitOn")
    )


class UnexpectedVehicleConnectionStatusError(Exception):
    pass



"""Basic unit tests for operations."""

from unittest.mock import AsyncMock, patch

import pytest
from aioresponses import aioresponses

from myskoda.anonymize import ACCESS_TOKEN, LOCATION, USER_ID, VIN
from myskoda.const import BASE_URL_SKODA
from myskoda.models.air_conditioning import (
    AirConditioning,
    AirConditioningAtUnlock,
    AirConditioningWithoutExternalPower,
    HeaterSource,
    SeatHeating,
    TargetTemperature,
    WindowHeating,
)
from myskoda.models.auxiliary_heating import AuxiliaryConfig, AuxiliaryHeating, AuxiliaryStartMode
from myskoda.models.charging import ChargeMode
from myskoda.models.departure import DepartureInfo
from myskoda.myskoda import MySkoda

from .conftest import FIXTURES_DIR, FakeMqttClientWrapper, create_aiomqtt_message


@pytest.mark.asyncio
async def test_stop_air_conditioning(
    responses: aioresponses, myskoda: MySkoda, fake_mqtt_client_wrapper: FakeMqttClientWrapper
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/stop"
    responses.post(url=url)

    future = myskoda.stop_air_conditioning(VIN)

    topic = f"{USER_ID}/{VIN}/operation-request/air-conditioning/start-stop-air-conditioning"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="stop-air-conditioning")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json=None,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("temperature", "expected"), [(21.5, "21.5"), (23.2, "23.0"), (10.01, "10.0")]
)
async def test_start_air_conditioning(
    responses: aioresponses,
    myskoda: MySkoda,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    temperature: float,
    expected: str,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/start"
    responses.post(url=url)

    future = myskoda.start_air_conditioning(VIN, temperature)

    topic = f"{USER_ID}/{VIN}/operation-request/air-conditioning/start-stop-air-conditioning"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="start-air-conditioning")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={
            "heaterSource": "ELECTRIC",
            "targetTemperature": {"temperatureValue": float(expected), "unitInCar": "CELSIUS"},
        },
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("temperature", "expected"), [(21.5, "21.5"), (23.2, "23.0"), (10.01, "10.0")]
)
async def test_set_target_temperature(
    responses: aioresponses,
    myskoda: MySkoda,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    temperature: float,
    expected: str,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/settings/target-temperature"
    responses.post(url=url)

    future = myskoda.set_target_temperature(VIN, temperature)

    topic = f"{USER_ID}/{VIN}/operation-request/air-conditioning/set-target-temperature"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="set-air-conditioning-target-temperature")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"temperatureValue": float(expected), "unitInCar": "CELSIUS"},
    )


@pytest.mark.asyncio
async def test_start_window_heating(
    responses: aioresponses, myskoda: MySkoda, fake_mqtt_client_wrapper: FakeMqttClientWrapper
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/start-window-heating"
    responses.post(url=url)

    future = myskoda.start_window_heating(VIN)

    topic = f"{USER_ID}/{VIN}/operation-request/air-conditioning/start-stop-window-heating"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="start-window-heating")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json=None,
    )


@pytest.mark.asyncio
async def test_stop_window_heating(
    responses: aioresponses, myskoda: MySkoda, fake_mqtt_client_wrapper: FakeMqttClientWrapper
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/stop-window-heating"
    responses.post(url=url)

    future = myskoda.stop_window_heating(VIN)

    topic = f"{USER_ID}/{VIN}/operation-request/air-conditioning/start-stop-window-heating"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="stop-window-heating")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json=None,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("limit", [50, 70, 90, 100])
async def test_set_charge_limit(
    responses: aioresponses,
    myskoda: MySkoda,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    limit: int,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/charging/{VIN}/set-charge-limit"
    responses.put(url=url)

    future = myskoda.set_charge_limit(VIN, limit)

    topic = f"{USER_ID}/{VIN}/operation-request/charging/update-charge-limit"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="update-charge-limit")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="PUT",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"targetSOCInPercent": limit},
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("limit", [0, 20, 30, 50])
async def test_set_minimum_charge_limit(
    responses: aioresponses,
    myskoda: MySkoda,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    limit: int,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/vehicle-automatization/{VIN}/departure/timers/settings"
    responses.post(url=url)

    future = myskoda.set_minimum_charge_limit(VIN, limit)

    topic = f"{USER_ID}/{VIN}/operation-request/departure/update-minimal-soc"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="update-minimal-soc")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"minimumBatteryStateOfChargeInPercent": limit},
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(("enabled", "expected"), [(True, "ACTIVATED"), (False, "DEACTIVATED")])
async def test_set_battery_care_mode(
    responses: aioresponses,
    myskoda: MySkoda,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    enabled: bool,
    expected: str,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/charging/{VIN}/set-care-mode"
    responses.put(url=url)

    future = myskoda.set_battery_care_mode(VIN, enabled)

    topic = f"{USER_ID}/{VIN}/operation-request/charging/update-care-mode"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="update-care-mode")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="PUT",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"chargingCareMode": expected},
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(("enabled", "expected"), [(True, "PERMANENT"), (False, "OFF")])
async def test_set_auto_unlock_plug(
    responses: aioresponses,
    myskoda: MySkoda,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    enabled: bool,
    expected: str,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/charging/{VIN}/set-auto-unlock-plug"
    responses.put(url=url)

    future = myskoda.set_auto_unlock_plug(VIN, enabled)

    topic = f"{USER_ID}/{VIN}/operation-request/charging/update-auto-unlock-plug"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="update-auto-unlock-plug")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="PUT",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"autoUnlockPlug": expected},
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(("reduced", "expected"), [(True, "REDUCED"), (False, "MAXIMUM")])
async def test_set_reduced_current_limit(
    responses: aioresponses,
    myskoda: MySkoda,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    reduced: bool,
    expected: str,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/charging/{VIN}/set-charging-current"
    responses.put(url=url)

    future = myskoda.set_reduced_current_limit(VIN, reduced)

    topic = f"{USER_ID}/{VIN}/operation-request/charging/update-charging-current"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="update-charging-current")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="PUT",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"chargingCurrent": expected},
    )


@pytest.mark.asyncio
async def test_start_charging(
    responses: aioresponses, myskoda: MySkoda, fake_mqtt_client_wrapper: FakeMqttClientWrapper
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/charging/{VIN}/start"
    responses.post(url=url)

    future = myskoda.start_charging(VIN)

    topic = f"{USER_ID}/{VIN}/operation-request/charging/start-stop-charging"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="start-charging")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json=None,
    )


@pytest.mark.asyncio
async def test_stop_charging(
    responses: aioresponses, myskoda: MySkoda, fake_mqtt_client_wrapper: FakeMqttClientWrapper
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/charging/{VIN}/stop"
    responses.post(url=url)

    future = myskoda.stop_charging(VIN)

    topic = f"{USER_ID}/{VIN}/operation-request/charging/start-stop-charging"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="stop-charging")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json=None,
    )


@pytest.mark.asyncio
async def test_wakeup(
    responses: aioresponses, myskoda: MySkoda, fake_mqtt_client_wrapper: FakeMqttClientWrapper
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/vehicle-wakeup/{VIN}?applyRequestLimiter=true"
    responses.post(url=url)

    future = myskoda.wakeup(VIN)

    topic = f"{USER_ID}/{VIN}/operation-request/vehicle-wakeup/wakeup"
    fake_mqtt_client_wrapper.set_messages([create_aiomqtt_message(topic=topic, operation="wakeup")])

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json=None,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("mode", ChargeMode)
async def test_set_charge_mode(
    responses: aioresponses,
    myskoda: MySkoda,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    mode: ChargeMode,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/charging/{VIN}/set-charge-mode"
    responses.post(url=url)

    future = myskoda.set_charge_mode(VIN, mode)

    topic = f"{USER_ID}/{VIN}/operation-request/charging/update-charge-mode"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="update-charge-mode")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"chargeMode": mode.value},
    )


@pytest.mark.asyncio
async def test_honk_and_flash(
    responses: aioresponses,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    myskoda: MySkoda,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/vehicle-access/{VIN}/honk-and-flash"
    responses.post(url=url)

    lat = LOCATION["latitude"]
    lng = LOCATION["longitude"]

    responses.get(
        url=f"{BASE_URL_SKODA}/api/v1/maps/positions?vin={VIN}",
        body=(FIXTURES_DIR / "enyaq" / "positions.json").read_text(),
    )

    future = myskoda.honk_flash(VIN)

    topic = f"{USER_ID}/{VIN}/operation-request/vehicle-access/honk-and-flash"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="start-honk")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"mode": "HONK_AND_FLASH", "vehiclePosition": {"latitude": lat, "longitude": lng}},
    )


@pytest.mark.asyncio
async def test_flash(
    responses: aioresponses,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    myskoda: MySkoda,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/vehicle-access/{VIN}/honk-and-flash"
    responses.post(url=url)

    lat = LOCATION["latitude"]
    lng = LOCATION["longitude"]

    responses.get(
        url=f"{BASE_URL_SKODA}/api/v1/maps/positions?vin={VIN}",
        body=(FIXTURES_DIR / "enyaq" / "positions.json").read_text(),
    )

    future = myskoda.flash(VIN)

    topic = f"{USER_ID}/{VIN}/operation-request/vehicle-access/honk-and-flash"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="start-flash")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"mode": "FLASH", "vehiclePosition": {"latitude": lat, "longitude": lng}},
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("spin", ["1234", "4321"])
async def test_lock(
    responses: aioresponses,
    myskoda: MySkoda,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    spin: str,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/vehicle-access/{VIN}/lock"
    responses.post(url=url)

    future = myskoda.lock(VIN, spin)

    topic = f"{USER_ID}/{VIN}/operation-request/vehicle-access/lock-vehicle"
    fake_mqtt_client_wrapper.set_messages([create_aiomqtt_message(topic=topic, operation="lock")])

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"currentSpin": spin},
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("spin", ["1234", "4321"])
async def test_unlock(
    responses: aioresponses,
    myskoda: MySkoda,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    spin: str,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/vehicle-access/{VIN}/unlock"
    responses.post(url=url)

    future = myskoda.unlock(VIN, spin)

    topic = f"{USER_ID}/{VIN}/operation-request/vehicle-access/lock-vehicle"
    fake_mqtt_client_wrapper.set_messages([create_aiomqtt_message(topic=topic, operation="unlock")])

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"currentSpin": spin},
    )


@pytest.mark.asyncio
async def test_stop_auxiliary_heater(
    responses: aioresponses, myskoda: MySkoda, fake_mqtt_client_wrapper: FakeMqttClientWrapper
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/auxiliary-heating/stop"
    responses.post(url=url)

    future = myskoda.stop_auxiliary_heating(VIN)

    topic = f"{USER_ID}/{VIN}/operation-request/auxiliary-heating/start-stop-auxiliary-heating"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="stop-auxiliary-heating")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json=None,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("spin", "config", "expected"),
    [
        ("1234", AuxiliaryConfig(TargetTemperature(temperature_value=22.3)), "22.5"),
        ("1234", AuxiliaryConfig(duration_in_seconds=600), "600"),
        ("1234", AuxiliaryConfig(start_mode=AuxiliaryStartMode.HEATING), "HEATING"),
        ("1234", AuxiliaryConfig(heater_source=HeaterSource.AUTOMATIC), "AUTOMATIC"),
    ],
)
async def test_start_auxiliary_heater(  # noqa: PLR0913
    responses: aioresponses,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    myskoda: MySkoda,
    spin: str,
    config: AuxiliaryConfig,
    expected: str,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/auxiliary-heating/start"
    responses.post(url=url)

    future = myskoda.start_auxiliary_heating(VIN, spin, config)

    topic = f"{USER_ID}/{VIN}/operation-request/auxiliary-heating/start-stop-auxiliary-heating"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="start-auxiliary-heating")]
    )

    json_data: dict[str, object] = {"spin": spin}
    if config is not None:
        if config.target_temperature is not None:
            json_data["targetTemperature"] = {
                "temperatureValue": float(expected),
                "unitInCar": "CELSIUS",
            }
        if config.duration_in_seconds is not None:
            json_data["durationInSeconds"] = int(expected)
        if config.heater_source is not None:
            json_data["heaterSource"] = expected
        if config.start_mode is not None:
            json_data["startMode"] = expected

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json=json_data,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("settings", "expected"),
    [
        (
            AirConditioningWithoutExternalPower(
                air_conditioning_without_external_power_enabled=True
            ),
            True,
        ),
        (
            AirConditioningWithoutExternalPower(
                air_conditioning_without_external_power_enabled=False
            ),
            False,
        ),
    ],
)
async def test_set_ac_without_external_power(
    responses: aioresponses,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    myskoda: MySkoda,
    settings: AirConditioningWithoutExternalPower,
    expected: bool,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/settings/ac-without-external-power"
    responses.post(url=url)

    future = myskoda.set_ac_without_external_power(VIN, settings)

    topic = (
        f"{USER_ID}/{VIN}/operation-request/"
        "air-conditioning/set-air-conditioning-without-external-power"
    )
    fake_mqtt_client_wrapper.set_messages(
        [
            create_aiomqtt_message(
                topic=topic, operation="set-air-conditioning-without-external-power"
            )
        ]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"airConditioningWithoutExternalPowerEnabled": expected},
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("settings", "expected"),
    [
        (AirConditioningAtUnlock(air_conditioning_at_unlock_enabled=True), True),
        (AirConditioningAtUnlock(air_conditioning_at_unlock_enabled=False), False),
    ],
)
async def test_set_ac_at_unlock(
    responses: aioresponses,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    myskoda: MySkoda,
    settings: AirConditioningAtUnlock,
    expected: bool,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/settings/ac-at-unlock"
    responses.post(url=url)

    future = myskoda.set_ac_at_unlock(VIN, settings)

    topic = f"{USER_ID}/{VIN}/operation-request/air-conditioning/set-air-conditioning-at-unlock"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="set-air-conditioning-at-unlock")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"airConditioningAtUnlockEnabled": expected},
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("settings", "expected"),
    [
        (WindowHeating(window_heating_enabled=True), True),
        (WindowHeating(window_heating_enabled=False), False),
    ],
)
async def test_set_windows_heating(
    responses: aioresponses,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    myskoda: MySkoda,
    settings: WindowHeating,
    expected: bool,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/settings/windows-heating"
    responses.post(url=url)

    future = myskoda.set_windows_heating(VIN, settings)

    topic = f"{USER_ID}/{VIN}/operation-request/air-conditioning/windows-heating"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="windows-heating")]
    )

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json={"windowHeatingEnabled": expected},
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ("settings", "expected"),
    [
        (SeatHeating(front_left=True), True),
        (SeatHeating(front_right=True), True),
        (SeatHeating(front_left=False), False),
    ],
)
async def test_set_seats_heating(
    responses: aioresponses,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    myskoda: MySkoda,
    settings: SeatHeating,
    expected: bool,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/settings/seats-heating"
    responses.post(url=url)

    future = myskoda.set_seats_heating(VIN, settings)

    topic = f"{USER_ID}/{VIN}/operation-request/air-conditioning/set-air-conditioning-seats-heating"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="set-air-conditioning-seats-heating")]
    )

    json_data: dict[str, object] = {}
    if settings is not None:
        if settings.front_left is not None:
            json_data["frontLeft"] = expected
        if settings.front_right is not None:
            json_data["frontRight"] = expected

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json=json_data,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(("timer_id", "enabled"), [(1, True), (2, False)])
async def test_set_departure_timer(
    responses: aioresponses,
    myskoda: MySkoda,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    timer_id: int,
    enabled: bool,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v1/vehicle-automatization/{VIN}/departure/timers"
    responses.post(url=url)

    departure_info_json = FIXTURES_DIR.joinpath("other/departure-timers.json").read_text()
    departure_info = DepartureInfo.from_json(departure_info_json)

    selected_timer = (
        next((timer for timer in departure_info.timers if timer.id == timer_id), None)
        if departure_info.timers
        else None
    )
    assert selected_timer is not None

    with patch("aiohttp.ClientSession._request", new_callable=AsyncMock) as mock_request:
        mock_request.return_value = AsyncMock(status=200)
        selected_timer.enabled = enabled
        future = myskoda.set_departure_timer(VIN, selected_timer)

        topic = f"{USER_ID}/{VIN}/operation-request/departure/update-departure-timers"
        fake_mqtt_client_wrapper.set_messages(
            [create_aiomqtt_message(topic=topic, operation="update-departure-timers")]
        )

        await future

        # Extract and assert the captured request body
        assert mock_request.called
        request_args, request_kwargs = mock_request.call_args
        body = request_kwargs.get("data") or request_kwargs.get("json")
        assert body is not None
        # check only the timer as deviceDateTime can't be verified
        assert body["timers"][0] == selected_timer.to_dict()


@pytest.mark.asyncio
@pytest.mark.parametrize(("timer_id", "enabled"), [(1, True), (2, False)])
async def test_set_ac_timer(
    responses: aioresponses,
    myskoda: MySkoda,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    timer_id: int,
    enabled: bool,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/timers"
    responses.post(url=url)

    ac_info_json = FIXTURES_DIR.joinpath("other/air-conditioning-idle.json").read_text()
    ac_info = AirConditioning.from_json(ac_info_json)

    selected_timer = (
        next((timer for timer in ac_info.timers if timer.id == timer_id), None)
        if ac_info.timers
        else None
    )
    assert selected_timer is not None

    selected_timer.enabled = enabled
    future = myskoda.set_ac_timer(VIN, selected_timer)

    topic = f"{USER_ID}/{VIN}/operation-request/air-conditioning/set-air-conditioning-timers"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="set-air-conditioning-timers")]
    )

    json_data = {"timers": [selected_timer.to_dict(by_alias=True)]}

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json=json_data,
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(("timer_id", "enabled", "spin"), [(1, True, "1234"), (2, False, "4321")])
async def test_set_auxiliary_heating_timer(  # noqa: PLR0913
    responses: aioresponses,
    fake_mqtt_client_wrapper: FakeMqttClientWrapper,
    myskoda: MySkoda,
    timer_id: int,
    enabled: bool,
    spin: str,
) -> None:
    url = f"{BASE_URL_SKODA}/api/v2/air-conditioning/{VIN}/auxiliary-heating/timers"
    responses.post(url=url)

    aux_info_json = FIXTURES_DIR.joinpath("other/auxiliary-heating-idle.json").read_text()
    aux_info = AuxiliaryHeating.from_json(aux_info_json)

    selected_timer = (
        next((timer for timer in aux_info.timers if timer.id == timer_id), None)
        if aux_info.timers
        else None
    )
    assert selected_timer is not None

    selected_timer.enabled = enabled
    future = myskoda.set_auxiliary_heating_timer(VIN, selected_timer, spin)

    topic = f"{USER_ID}/{VIN}/operation-request/air-conditioning/set-air-conditioning-timers"
    fake_mqtt_client_wrapper.set_messages(
        [create_aiomqtt_message(topic=topic, operation="set-air-conditioning-timers")]
    )

    json_data = {"spin": spin, "timers": [selected_timer.to_dict(by_alias=True)]}

    await future
    responses.assert_called_with(
        url=url,
        method="POST",
        headers={"authorization": f"Bearer {ACCESS_TOKEN}"},
        json=json_data,
    )



"""Unit tests for utilities."""

import asyncio

import pytest

from myskoda.utils import async_debounce


@pytest.mark.asyncio
async def test_async_debounce() -> None:
    """Test the async_debounce decorator.

    Debounced function:
    - Only executes once when called multiple times within 'wait'.
    - Executes twice when called two times with a delay > 'wait'.
    """
    test_val = 0

    @async_debounce(wait=0.2)
    async def increment() -> None:
        nonlocal test_val
        test_val += 1

    await increment()
    await increment()
    await asyncio.sleep(0.1)
    await increment()
    await asyncio.sleep(0.3)
    assert test_val == 1

    await increment()
    await asyncio.sleep(0.3)
    await increment()
    await asyncio.sleep(0.3)
    assert test_val == 3  # noqa: PLR2004


@pytest.mark.asyncio
async def test_async_debounce_immediate() -> None:
    """Test the async_debounce decorator when immediate is True.

    Debounced function:
    - Executes immediatally the first time called.
    - Subsequent calls are executed (once) after wait time.
    """
    test_val = 0

    @async_debounce(wait=0.2, immediate=True)
    async def increment() -> None:
        nonlocal test_val
        test_val += 1

    await increment()
    assert test_val == 1
    await increment()
    await increment()
    await asyncio.sleep(0.3)
    assert test_val == 2  # noqa: PLR2004


@pytest.mark.asyncio
async def test_async_debounce_immediate_noqueue() -> None:
    """Test the async_debounce decorator when immediate is True and queue is False.

    Debounced function:
    - Executes immediatally the first time called.
    - Subsequent calls are executed (once) after wait time.
    """
    test_val = 0

    @async_debounce(wait=0.2, immediate=True, queue=False)
    async def increment() -> None:
        nonlocal test_val
        test_val += 1

    await increment()
    assert test_val == 1
    await increment()
    await increment()
    await asyncio.sleep(0.3)
    assert test_val == 1



"""Main entry point for the MySkoda library.

This class provides all methods to operate on the API and MQTT broker.

Example:
    async with aiohttp.ClientSession() as client:
        myskoda = MySkoda(session)
        await myskoda.connect("username", "password")
        for vin in await myskoda.list_vehicle_vins():
            print(vin)

All get_ methods will always fetch new data from the API and return that.

All refresh_ methods are debounced or otherwise rate limited. They (eventually) update local
attributes and don't return anything. Instead they trigger callbacks which clients can register
for using the subscribe_updates method.

MQTT event callbacks can also be subscribed for using the subscribe_events method.

"""

import asyncio
import logging
from collections import defaultdict
from collections.abc import Callable, Coroutine
from datetime import UTC, datetime, timedelta
from ssl import SSLContext
from traceback import format_exc
from types import SimpleNamespace
from typing import Any

from aiohttp import ClientSession, TraceConfig, TraceRequestEndParams

from myskoda.anonymize import anonymize_url
from myskoda.models.fixtures import (
    Endpoint,
    Fixture,
    FixtureReportGet,
    FixtureReportType,
    FixtureVehicle,
    create_fixture_vehicle,
)

from .__version__ import __version__ as version
from .auth.authorization import Authorization
from .const import (
    BASE_URL_SKODA,
    CACHE_USER_ENDPOINT_IN_HOURS,
    CACHE_VEHICLE_HEALTH_IN_HOURS,
    CLIENT_ID,
    MQTT_OPERATION_TIMEOUT,
    OPERATION_REFRESH_DELAY_SECONDS,
    REDIRECT_URI,
)
from .models.air_conditioning import (
    AirConditioning,
    AirConditioningAtUnlock,
    AirConditioningTimer,
    AirConditioningWithoutExternalPower,
    SeatHeating,
    WindowHeating,
)
from .models.auxiliary_heating import (
    AuxiliaryConfig,
    AuxiliaryHeating,
    AuxiliaryHeatingTimer,
)
from .models.charging import ChargeMode, Charging, ChargingStatus
from .models.chargingprofiles import ChargingProfiles
from .models.common import Vin
from .models.departure import DepartureInfo, DepartureTimer
from .models.driving_range import DrivingRange, EngineType
from .models.event import (
    BaseEvent,
    OperationEvent,
    OperationName,
    OperationStatus,
    ServiceEventAccess,
    ServiceEventAirConditioning,
    ServiceEventChangeSoc,
    ServiceEventChangeSocData,
    ServiceEventCharging,
    ServiceEventDeparture,
    ServiceEventOdometer,
)
from .models.health import Health
from .models.info import CapabilityId, Info
from .models.maintenance import Maintenance, MaintenanceReport
from .models.position import Positions
from .models.spin import Spin
from .models.status import Status
from .models.trip_statistics import TripStatistics
from .models.user import User
from .models.vehicle_connection_status import VehicleConnectionStatus
from .mqtt import MySkodaMqttClient
from .rest_api import GetEndpointResult, RestApi
from .utils import async_debounce
from .vehicle import Vehicle

_LOGGER = logging.getLogger(__name__)

background_tasks = set()


class MqttDisabledError(Exception):
    """MQTT was not enabled."""


class UnsupportedEndpointError(Exception):
    """Endpoint not implemented."""


class UnknownVinError(Exception):
    """Requested Vin not found."""

    def __init__(self, vin: str) -> None:
        super().__init__(f"Vehicle with VIN {vin} not found")


async def trace_response(
    _session: ClientSession,
    _trace_config_ctx: SimpleNamespace,
    params: TraceRequestEndParams,
) -> None:
    """Log response details. Used in aiohttp.TraceConfig."""
    resp_text = await params.response.text()
    _LOGGER.debug(
        "Trace: %s %s - response: %s (%s bytes) %s",
        params.method,
        str(params.url)[:60],
        params.response.status,
        params.response.content_length,
        resp_text[:5000],
    )


TRACE_CONFIG = TraceConfig()
TRACE_CONFIG.on_request_end.append(trace_response)


class MySkodaAuthorization(Authorization):
    client_id: str = CLIENT_ID  #  pyright: ignore[reportIncompatibleMethodOverride]
    redirect_uri: str = REDIRECT_URI  #  pyright: ignore[reportIncompatibleMethodOverride]
    base_url: str = BASE_URL_SKODA  #  pyright: ignore[reportIncompatibleMethodOverride]


class MySkoda:
    session: ClientSession
    rest_api: RestApi
    mqtt: MySkodaMqttClient | None = None
    authorization: MySkodaAuthorization
    ssl_context: SSLContext | None = None
    user: User | None = None
    _vehicles: dict[Vin, Vehicle]
    _callbacks: dict[Vin, list[Callable[[], Coroutine[Any, Any, None]]]]

    def __init__(
        self,
        session: ClientSession,
        ssl_context: SSLContext | None = None,
        mqtt_enabled: bool = True,
    ) -> None:
        self._callbacks = defaultdict(list)
        self._vehicles: dict[Vin, Vehicle] = {}
        self.session = session
        self.authorization = MySkodaAuthorization(session)
        self.rest_api = RestApi(self.session, self.authorization)
        self.ssl_context = ssl_context
        if mqtt_enabled:
            self.mqtt = self._create_mqtt_client()

    async def enable_mqtt(self) -> None:
        """If MQTT was not enabled when initializing MySkoda, enable it manually and connect."""
        if self.mqtt is not None:
            return
        self.mqtt = self._create_mqtt_client()
        self.user = await self.get_user()
        vehicles = await self.list_vehicle_vins()
        await self.mqtt.connect(self.user.id, vehicles)

    async def connect(self, email: str, password: str) -> None:
        """Authenticate on the rest api and connect to the MQTT broker."""
        await self.authorization.authorize(email, password)
        _LOGGER.debug("IDK Authorization was successful.")

        if self.mqtt:
            user = await self.get_user()
            vehicles = await self.list_vehicle_vins()
            await self.mqtt.connect(user.id, vehicles)
        _LOGGER.debug("MySkoda ready.")

    async def disconnect(self) -> None:
        """Disconnect from the MQTT broker."""
        if self.mqtt:
            await self.mqtt.disconnect()

    def subscribe_events(self, callback: Callable[[BaseEvent], Coroutine[Any, Any, None]]) -> None:
        """Listen for events emitted by MySkoda's MQTT broker."""
        if self.mqtt is None:
            raise MqttDisabledError
        self.mqtt.subscribe(callback=callback)

    def subscribe(self, callback: Callable[[BaseEvent], Coroutine[Any, Any, None]]) -> None:
        """See subscribe_events. For backwards compatibility."""
        _LOGGER.warning(
            "The subscribe() method is deprecated and will be removed, use subscribe_events"
        )
        self.subscribe_events(callback=callback)

    def subscribe_updates(
        self, vin: Vin, callback: Callable[[], Coroutine[Any, Any, None]]
    ) -> None:
        """Subscribe a callback function to be called when Vehicle data is updated."""
        self._callbacks[vin].append(callback)

    async def verify_spin(self, spin: str, anonymize: bool = False) -> Spin:
        """Verify S-PIN."""
        return (await self.rest_api.verify_spin(spin, anonymize=anonymize)).result

    async def list_vehicle_vins(self) -> list[str]:
        """List all vehicles by their vins."""
        garage = (await self.rest_api.get_garage()).result
        if garage.vehicles is None:
            return []
        return [vehicle.vin for vehicle in garage.vehicles]

    def vehicle(self, vin: Vin) -> Vehicle:
        """Return the currently cached vehicle."""
        if vin in self._vehicles:
            return self._vehicles[vin]
        raise UnknownVinError(vin)

    async def get_vehicle(
        self, vin: Vin, excluded_capabilities: list[CapabilityId] | None = None
    ) -> Vehicle:
        """Load and return a full vehicle based on its capabilities."""
        capabilities = [
            CapabilityId.AIR_CONDITIONING,
            CapabilityId.AUXILIARY_HEATING,
            CapabilityId.CHARGING,
            CapabilityId.PARKING_POSITION,
            CapabilityId.STATE,
            CapabilityId.TRIP_STATISTICS,
            CapabilityId.VEHICLE_HEALTH_INSPECTION,
            CapabilityId.DEPARTURE_TIMERS,
            CapabilityId.READINESS,
        ]

        if excluded_capabilities:
            capabilities = [c for c in capabilities if c not in excluded_capabilities]

        return await self.get_partial_vehicle(vin, capabilities)

    async def get_partial_vehicle(self, vin: Vin, capabilities: list[CapabilityId]) -> Vehicle:
        """Load and return a partial vehicle, based on list of capabilities."""
        info = await self.get_info(vin)
        maintenance = await self.get_maintenance(vin)

        if vin in self._vehicles:
            self._vehicles[vin].info = info
            self._vehicles[vin].maintenance = maintenance
        else:
            self._vehicles[vin] = Vehicle(info=info, maintenance=maintenance)

        for capa in capabilities:
            if info.is_capability_available(capa):
                await self._request_capability_data(vin, capa)

        return self.vehicle(vin)

    async def get_auth_token(self) -> str:
        """Retrieve the main access token for the IDK session."""
        return await self.rest_api.authorization.get_access_token()

    async def get_user(self, anonymize: bool = False) -> User:
        """Retrieve user information about logged in user."""
        return (await self.rest_api.get_user(anonymize=anonymize)).result

    async def get_info(self, vin: Vin, anonymize: bool = False) -> Info:
        """Retrieve the basic vehicle information for the specified vehicle."""
        return (await self.rest_api.get_info(vin, anonymize=anonymize)).result

    async def get_charging(self, vin: Vin, anonymize: bool = False) -> Charging:
        """Retrieve information related to charging for the specified vehicle."""
        return (await self.rest_api.get_charging(vin, anonymize=anonymize)).result

    async def get_charging_profiles(self, vin: Vin, anonymize: bool = False) -> ChargingProfiles:
        """Retrieve information related to charging profiles for the specified vehicle."""
        return (await self.rest_api.get_charging_profiles(vin, anonymize=anonymize)).result

    async def get_status(self, vin: Vin, anonymize: bool = False) -> Status:
        """Retrieve the current status for the specified vehicle."""
        return (await self.rest_api.get_status(vin, anonymize=anonymize)).result

    async def get_air_conditioning(self, vin: Vin, anonymize: bool = False) -> AirConditioning:
        """Retrieve the current air conditioning status for the specified vehicle."""
        return (await self.rest_api.get_air_conditioning(vin, anonymize=anonymize)).result

    async def get_auxiliary_heating(self, vin: Vin, anonymize: bool = False) -> AuxiliaryHeating:
        """Retrieve the current auxiliary heating status for the specified vehicle."""
        return (await self.rest_api.get_auxiliary_heating(vin, anonymize=anonymize)).result

    async def get_positions(self, vin: Vin, anonymize: bool = False) -> Positions:
        """Retrieve the current position for the specified vehicle."""
        return (await self.rest_api.get_positions(vin, anonymize=anonymize)).result

    async def get_driving_range(self, vin: Vin, anonymize: bool = False) -> DrivingRange:
        """Retrieve estimated driving range for combustion vehicles."""
        return (await self.rest_api.get_driving_range(vin, anonymize=anonymize)).result

    async def get_trip_statistics(self, vin: Vin, anonymize: bool = False) -> TripStatistics:
        """Retrieve statistics about past trips."""
        return (await self.rest_api.get_trip_statistics(vin, anonymize=anonymize)).result

    async def get_maintenance(self, vin: Vin, anonymize: bool = False) -> Maintenance:
        """Retrieve maintenance report, settings and history."""
        return (await self.rest_api.get_maintenance(vin, anonymize=anonymize)).result

    async def get_maintenance_report(self, vin: Vin, anonymize: bool = False) -> MaintenanceReport:
        """Retrieve maintenance report only."""
        return (await self.rest_api.get_maintenance_report(vin, anonymize=anonymize)).result

    async def get_health(self, vin: Vin, anonymize: bool = False) -> Health:
        """Retrieve health information for the specified vehicle."""
        return (await self.rest_api.get_health(vin, anonymize=anonymize)).result

    async def get_departure_timers(self, vin: Vin, anonymize: bool = False) -> DepartureInfo:
        """Retrieve departure timers for the specified vehicle."""
        return (await self.rest_api.get_departure_timers(vin, anonymize=anonymize)).result

    async def get_connection_status(
        self, vin: Vin, anonymize: bool = False
    ) -> VehicleConnectionStatus:
        """Retrieve vehicle connection status for the specified vehicle."""
        return (await self.rest_api.get_vehicle_connection_status(vin, anonymize=anonymize)).result

    async def start_charging(self, vin: Vin) -> None:
        """Start charging the car."""
        future = self._wait_for_operation(OperationName.START_CHARGING)
        await self.rest_api.start_charging(vin)
        await future

    async def stop_charging(self, vin: Vin) -> None:
        """Stop the car from charging."""
        future = self._wait_for_operation(OperationName.STOP_CHARGING)
        await self.rest_api.stop_charging(vin)
        await future

    async def set_charge_mode(self, vin: Vin, mode: ChargeMode) -> None:
        """Set the charge mode."""
        future = self._wait_for_operation(OperationName.UPDATE_CHARGE_MODE)
        await self.rest_api.set_charge_mode(vin, mode=mode)
        await future

    async def honk_flash(self, vin: Vin) -> None:
        """Honk and flash."""
        future = self._wait_for_operation(OperationName.START_HONK)
        await self.rest_api.honk_flash(vin, (await self.get_positions(vin)).positions)
        await future

    async def flash(self, vin: Vin) -> None:
        """Flash lights."""
        future = self._wait_for_operation(OperationName.START_FLASH)
        await self.rest_api.flash(vin, (await self.get_positions(vin)).positions)
        await future

    async def wakeup(self, vin: Vin) -> None:
        """Wake the vehicle up. Can be called maximum three times a day."""
        future = self._wait_for_operation(OperationName.WAKEUP)
        await self.rest_api.wakeup(vin)
        await future

    async def set_reduced_current_limit(self, vin: Vin, reduced: bool) -> None:
        """Enable reducing the current limit by which the car is charged."""
        future = self._wait_for_operation(OperationName.UPDATE_CHARGING_CURRENT)
        await self.rest_api.set_reduced_current_limit(vin, reduced=reduced)
        await future

    async def set_battery_care_mode(self, vin: Vin, enabled: bool) -> None:
        """Enable or disable the battery care mode."""
        future = self._wait_for_operation(OperationName.UPDATE_CARE_MODE)
        await self.rest_api.set_battery_care_mode(vin, enabled)
        await future

    async def set_auto_unlock_plug(self, vin: Vin, enabled: bool) -> None:
        """Enable or disable auto unlock plug when charged."""
        future = self._wait_for_operation(OperationName.UPDATE_AUTO_UNLOCK_PLUG)
        await self.rest_api.set_auto_unlock_plug(vin, enabled)
        await future

    async def set_charge_limit(self, vin: Vin, limit: int) -> None:
        """Set the maximum charge limit in percent."""
        future = self._wait_for_operation(OperationName.UPDATE_CHARGE_LIMIT)
        await self.rest_api.set_charge_limit(vin, limit)
        await future

    async def set_minimum_charge_limit(self, vin: Vin, limit: int) -> None:
        """Set minimum battery SoC in percent for departure timer."""
        future = self._wait_for_operation(OperationName.UPDATE_MINIMAL_SOC)
        await self.rest_api.set_minimum_charge_limit(vin, limit)
        await future

    async def stop_window_heating(self, vin: Vin) -> None:
        """Stop heating both the front and rear window."""
        future = self._wait_for_operation(OperationName.STOP_WINDOW_HEATING)
        await self.rest_api.stop_window_heating(vin)
        await future

    async def start_window_heating(self, vin: Vin) -> None:
        """Start heating both the front and rear window."""
        future = self._wait_for_operation(OperationName.START_WINDOW_HEATING)
        await self.rest_api.start_window_heating(vin)
        await future

    async def set_ac_without_external_power(
        self, vin: Vin, settings: AirConditioningWithoutExternalPower
    ) -> None:
        """Enable or disable AC without external power."""
        future = self._wait_for_operation(OperationName.SET_AIR_CONDITIONING_WITHOUT_EXTERNAL_POWER)
        await self.rest_api.set_ac_without_external_power(vin, settings)
        await future

    async def set_ac_at_unlock(self, vin: Vin, settings: AirConditioningAtUnlock) -> None:
        """Enable or disable AC at unlock."""
        future = self._wait_for_operation(OperationName.SET_AIR_CONDITIONING_AT_UNLOCK)
        await self.rest_api.set_ac_at_unlock(vin, settings)
        await future

    async def set_windows_heating(self, vin: Vin, settings: WindowHeating) -> None:
        """Enable or disable windows heating with AC."""
        future = self._wait_for_operation(OperationName.WINDOWS_HEATING)
        await self.rest_api.set_windows_heating(vin, settings)
        await future

    async def set_seats_heating(self, vin: Vin, settings: SeatHeating) -> None:
        """Enable or disable seats heating with AC."""
        future = self._wait_for_operation(OperationName.SET_AIR_CONDITIONING_SEATS_HEATING)
        await self.rest_api.set_seats_heating(vin, settings)
        await future

    async def set_target_temperature(self, vin: Vin, temperature: float) -> None:
        """Set the air conditioning's target temperature in °C."""
        future = self._wait_for_operation(OperationName.SET_AIR_CONDITIONING_TARGET_TEMPERATURE)
        await self.rest_api.set_target_temperature(vin, temperature)
        await future

    async def start_air_conditioning(self, vin: Vin, temperature: float) -> None:
        """Start the air conditioning with the provided target temperature in °C."""
        future = self._wait_for_operation(OperationName.START_AIR_CONDITIONING)
        await self.rest_api.start_air_conditioning(vin, temperature)
        await future

    async def stop_air_conditioning(self, vin: Vin) -> None:
        """Stop the air conditioning."""
        future = self._wait_for_operation(OperationName.STOP_AIR_CONDITIONING)
        await self.rest_api.stop_air_conditioning(vin)
        await future

    async def start_ventilation(self, vin: Vin) -> None:
        """Start the ventilation."""
        future = self._wait_for_operation(OperationName.START_ACTIVE_VENTILATION)
        await self.rest_api.start_ventilation(vin)
        await future

    async def stop_ventilation(self, vin: Vin) -> None:
        """Start the ventilation."""
        future = self._wait_for_operation(OperationName.STOP_ACTIVE_VENTILATION)
        await self.rest_api.stop_ventilation(vin)
        await future

    async def start_auxiliary_heating(
        self, vin: Vin, spin: str, config: AuxiliaryConfig | None = None
    ) -> None:
        """Start the auxiliary heating with the provided configuration."""
        future = self._wait_for_operation(OperationName.START_AUXILIARY_HEATING)
        await self.rest_api.start_auxiliary_heating(vin, spin, config=config)
        await future

    async def stop_auxiliary_heating(self, vin: Vin) -> None:
        """Stop the auxiliary heating."""
        future = self._wait_for_operation(OperationName.STOP_AUXILIARY_HEATING)
        await self.rest_api.stop_auxiliary_heating(vin)
        await future

    async def set_ac_timer(self, vin: Vin, timer: AirConditioningTimer) -> None:
        """Send provided air-conditioning timer to the vehicle."""
        future = self._wait_for_operation(OperationName.SET_AIR_CONDITIONING_TIMERS)
        await self.rest_api.set_ac_timer(vin, timer)
        await future

    async def set_auxiliary_heating_timer(
        self, vin: Vin, timer: AuxiliaryHeatingTimer, spin: str
    ) -> None:
        """Send provided auxiliary heating timer to the vehicle."""
        future = self._wait_for_operation(OperationName.SET_AIR_CONDITIONING_TIMERS)
        await self.rest_api.set_auxiliary_heating_timer(vin, timer, spin)
        await future

    async def lock(self, vin: Vin, spin: str) -> None:
        """Lock the car."""
        future = self._wait_for_operation(OperationName.LOCK)
        await self.rest_api.lock(vin, spin)
        await future

    async def unlock(self, vin: Vin, spin: str) -> None:
        """Unlock the car."""
        future = self._wait_for_operation(OperationName.UNLOCK)
        await self.rest_api.unlock(vin, spin)
        await future

    async def set_departure_timer(self, vin: Vin, timer: DepartureTimer) -> None:
        """Send provided departure timer to the vehicle."""
        future = self._wait_for_operation(OperationName.UPDATE_DEPARTURE_TIMERS)
        await self.rest_api.set_departure_timer(vin, timer)
        await future

    async def refresh_user(self) -> None:
        """Refresh user data for the provided Vin."""
        if self.user and self.user.timestamp:
            cache_expiry_time = self.user.timestamp + timedelta(hours=CACHE_USER_ENDPOINT_IN_HOURS)

            if datetime.now(UTC) > cache_expiry_time:
                _LOGGER.debug("Refreshing user - cache expired at %s", self.user.timestamp)
                self.user = await self.get_user()
            else:
                _LOGGER.debug("Skipping user refresh - cache is still valid.")
        else:
            self.user = await self.get_user()

    @async_debounce(immediate=True)
    async def refresh_vehicle(self, vin: Vin, notify: bool = True) -> None:
        """Refresh all vehicle data for the provided Vin.

        Get health only when missing and every 24h.
        This avoids triggering battery protection, such as in Citigoe and Karoq.
        https://github.com/skodaconnect/homeassistant-myskoda/issues/468
        """
        excluded_capabilities = []
        if (vehicle := self._vehicles.get(vin)) and vehicle.health and vehicle.health.timestamp:
            cache_expiry = vehicle.health.timestamp + timedelta(hours=CACHE_VEHICLE_HEALTH_IN_HOURS)

            if datetime.now(UTC) > cache_expiry:
                _LOGGER.debug("Refreshing health - cache expired at %s", cache_expiry)
            else:
                _LOGGER.debug("Skipping health refresh - cache is still valid.")
                excluded_capabilities.append(CapabilityId.VEHICLE_HEALTH_INSPECTION)

        self._vehicles[vin] = await self.get_vehicle(vin, excluded_capabilities)

        if notify:
            self._notify_callbacks(vin)

    @async_debounce(immediate=True)
    async def refresh_info(self, vin: Vin, notify: bool = True) -> None:
        """Refresh info data for the provided Vin."""
        self._vehicles[vin].info = await self.get_info(vin)
        if notify:
            self._notify_callbacks(vin)

    @async_debounce(immediate=True)
    async def refresh_charging(self, vin: Vin, notify: bool = True) -> None:
        """Refresh charging data for the provided Vin."""
        self._vehicles[vin].charging = await self.get_charging(vin)
        if notify:
            self._notify_callbacks(vin)

    @async_debounce(immediate=True)
    async def refresh_status(self, vin: Vin, notify: bool = True) -> None:
        """Refresh status data for the provided Vin."""
        self._vehicles[vin].status = await self.get_status(vin)
        if notify:
            self._notify_callbacks(vin)

    @async_debounce(immediate=True)
    async def refresh_air_conditioning(self, vin: Vin, notify: bool = True) -> None:
        """Refresh air_conditioning data for the provided Vin."""
        self._vehicles[vin].air_conditioning = await self.get_air_conditioning(vin)
        if notify:
            self._notify_callbacks(vin)

    @async_debounce(immediate=True)
    async def refresh_auxiliary_heating(self, vin: Vin, notify: bool = True) -> None:
        """Refresh auxiliary_heating data for the provided Vin."""
        self._vehicles[vin].auxiliary_heating = await self.get_auxiliary_heating(vin)
        if notify:
            self._notify_callbacks(vin)

    @async_debounce(immediate=True)
    async def refresh_positions(self, vin: Vin, notify: bool = True) -> None:
        """Refresh positions data for the provided Vin."""
        self._vehicles[vin].positions = await self.get_positions(vin)
        if notify:
            self._notify_callbacks(vin)

    @async_debounce(immediate=True)
    async def refresh_driving_range(self, vin: Vin, notify: bool = True) -> None:
        """Refresh driving_range data for the provided Vin."""
        self._vehicles[vin].driving_range = await self.get_driving_range(vin)
        if notify:
            self._notify_callbacks(vin)

    @async_debounce(immediate=True)
    async def refresh_trip_statistics(self, vin: Vin, notify: bool = True) -> None:
        """Refresh trip_statistics data for the provided Vin."""
        self._vehicles[vin].trip_statistics = await self.get_trip_statistics(vin)
        if notify:
            self._notify_callbacks(vin)

    @async_debounce(immediate=True)
    async def refresh_maintenance(self, vin: Vin, notify: bool = True) -> None:
        """Refresh maintenance data for the provided Vin."""
        self._vehicles[vin].maintenance = await self.get_maintenance(vin)
        if notify:
            self._notify_callbacks(vin)

    @async_debounce(immediate=True)
    async def refresh_health(self, vin: Vin, notify: bool = True) -> None:
        """Refresh health data for the provided Vin."""
        self._vehicles[vin].health = await self.get_health(vin)
        if notify:
            self._notify_callbacks(vin)

    @async_debounce(immediate=True)
    async def refresh_departure_info(self, vin: Vin, notify: bool = True) -> None:
        """Refresh departure_info data for the provided Vin."""
        self._vehicles[vin].departure_info = await self.get_departure_timers(vin)
        if notify:
            self._notify_callbacks(vin)

    async def generate_fixture_report(
        self, vin: Vin, vehicle: FixtureVehicle, endpoint: Endpoint
    ) -> FixtureReportGet:
        """Generate a fixture report for the specified endpoint and vehicle."""
        try:
            result = await self.get_endpoint(vin, endpoint, anonymize=True)
        except Exception:  # noqa: BLE001
            return FixtureReportGet(
                type=FixtureReportType.GET,
                vehicle_id=vehicle.id,
                success=False,
                endpoint=endpoint,
                error=anonymize_url(format_exc()),
            )
        else:
            return FixtureReportGet(
                type=FixtureReportType.GET,
                vehicle_id=vehicle.id,
                raw=result.raw,
                success=True,
                url=result.url,
                endpoint=endpoint,
                result=result.result.to_dict(),
            )

    async def get_endpoint(
        self, vin: Vin, endpoint: Endpoint, anonymize: bool = False
    ) -> GetEndpointResult[Any]:
        """Invoke a get endpoint by endpoint enum."""
        # Mapping of endpoints to corresponding methods
        endpoint_method_map = {
            Endpoint.INFO: self.rest_api.get_info,
            Endpoint.STATUS: self.rest_api.get_status,
            Endpoint.AIR_CONDITIONING: self.rest_api.get_air_conditioning,
            Endpoint.AUXILIARY_HEATING: self.rest_api.get_auxiliary_heating,
            Endpoint.POSITIONS: self.rest_api.get_positions,
            Endpoint.HEALTH: self.rest_api.get_health,
            Endpoint.CHARGING: self.rest_api.get_charging,
            Endpoint.CHARGING_PROFILES: self.rest_api.get_charging_profiles,
            Endpoint.MAINTENANCE: self.rest_api.get_maintenance,
            Endpoint.DRIVING_RANGE: self.rest_api.get_driving_range,
            Endpoint.TRIP_STATISTICS: self.rest_api.get_trip_statistics,
            Endpoint.DEPARTURE_INFO: self.rest_api.get_departure_timers,
            Endpoint.VEHICLE_CONNECTION_STATUS: self.rest_api.get_vehicle_connection_status,
        }

        # Look up the method, or raise an error if unsupported
        method = endpoint_method_map.get(endpoint)
        if not method:
            error_message = f"Unsupported endpoint: {endpoint}"
            raise UnsupportedEndpointError(error_message)

        # Call the method and return the result
        return await method(vin, anonymize=anonymize)

    async def generate_get_fixture(
        self, name: str, description: str, vins: list[str], endpoint: Endpoint
    ) -> Fixture:
        """Generate a fixture for a get request."""
        vehicles = [
            (vin, create_fixture_vehicle(i, await self.get_info(vin))) for i, vin in enumerate(vins)
        ]

        endpoints = []
        if endpoint != Endpoint.ALL:
            endpoints = [endpoint]
        else:
            endpoints = filter(lambda ep: ep != Endpoint.ALL, Endpoint)

        reports = [
            await self.generate_fixture_report(vin, vehicle, endpoint)
            for (vin, vehicle) in vehicles
            for endpoint in endpoints
        ]

        return Fixture(
            name=name,
            description=description,
            generation_time=datetime.now(tz=UTC),
            vehicles=[vehicle for (_, vehicle) in vehicles],
            reports=reports,
            library_version=version,
        )

    async def _request_capability_data(self, vin: Vin, capa: CapabilityId) -> None:
        """Request specific capability data from MySkoda API."""
        capa_request_map = {
            CapabilityId.AIR_CONDITIONING: self._request_air_conditioning,
            CapabilityId.AUXILIARY_HEATING: self._request_auxiliary_heating,
            CapabilityId.CHARGING: self._request_charging,
            CapabilityId.PARKING_POSITION: self._request_positions,
            CapabilityId.STATE: self._request_state,
            CapabilityId.TRIP_STATISTICS: self._request_trip_statistics,
            CapabilityId.VEHICLE_HEALTH_INSPECTION: self._request_health,
            CapabilityId.DEPARTURE_TIMERS: self._request_departure_info,
            CapabilityId.READINESS: self._request_connection_status,
        }

        try:
            request_fn = capa_request_map.get(capa)
            if request_fn:
                await request_fn(vin)
        except Exception as err:  # noqa: BLE001
            _LOGGER.warning("Requesting %s failed: %s, continue", capa, err)

    async def _request_air_conditioning(self, vin: Vin) -> None:
        """Update state with air conditioning data."""
        self._vehicles[vin].air_conditioning = await self.get_air_conditioning(vin)

    async def _request_auxiliary_heating(self, vin: Vin) -> None:
        """Update state with auxiliary heating data."""
        self._vehicles[vin].auxiliary_heating = await self.get_auxiliary_heating(vin)

    async def _request_charging(self, vin: Vin) -> None:
        """Update state with charging data."""
        self._vehicles[vin].charging = await self.get_charging(vin)

    async def _request_positions(self, vin: Vin) -> None:
        """Update state with parking position data."""
        self._vehicles[vin].positions = await self.get_positions(vin)

    async def _request_state(self, vin: Vin) -> None:
        """Update state with state and driving range data."""
        self._vehicles[vin].status = await self.get_status(vin)
        self._vehicles[vin].driving_range = await self.get_driving_range(vin)

    async def _request_trip_statistics(self, vin: Vin) -> None:
        """Update state with trip statistics data."""
        self._vehicles[vin].trip_statistics = await self.get_trip_statistics(vin)

    async def _request_health(self, vin: Vin) -> None:
        """Update state with vehicle health inspection data."""
        self._vehicles[vin].health = await self.get_health(vin)

    async def _request_departure_info(self, vin: Vin) -> None:
        """Update state with departure timer data."""
        self._vehicles[vin].departure_info = await self.get_departure_timers(vin)

    async def _request_connection_status(self, vin: Vin) -> None:
        """Update state with connection status data."""
        self._vehicles[vin].connection_status = await self.get_connection_status(vin)

    async def _wait_for_operation(self, operation: OperationName) -> None:
        if self.mqtt is None:
            return
        try:
            async with asyncio.timeout(MQTT_OPERATION_TIMEOUT):
                await self.mqtt.wait_for_operation(operation)
        except TimeoutError:
            _LOGGER.warning("Timeout occurred while waiting for %s. Aborted.", operation)

    def _notify_callbacks(self, vin: Vin) -> None:
        """Execute registered callback functions for the vin."""
        for callback in self._callbacks.get(vin, []):
            result = callback()
            if result is not None:
                task = asyncio.create_task(result)
                background_tasks.add(task)
                task.add_done_callback(background_tasks.discard)

    def _create_mqtt_client(self) -> MySkodaMqttClient:
        mqtt = MySkodaMqttClient(authorization=self.authorization, ssl_context=self.ssl_context)
        mqtt.subscribe(self._on_mqtt_event)
        return mqtt

    async def _on_mqtt_event(self, event: BaseEvent) -> None:
        """Handle MQTT events.

        Update self._vehicles with data received in event and notify callbacks.
        """
        if event.vin not in self._vehicles:
            _LOGGER.debug("Received event for unknown VIN %s", event)
            return

        if isinstance(event, OperationEvent):
            await self._process_operation_event(event)
        elif isinstance(event, ServiceEventChangeSoc):
            await self._process_charging_event(event)
        elif isinstance(event, ServiceEventCharging):
            await self.refresh_charging(event.vin)
        elif isinstance(event, ServiceEventAccess):
            await self.refresh_vehicle(event.vin)
        elif isinstance(event, ServiceEventAirConditioning):
            await self.refresh_air_conditioning(event.vin)
        elif isinstance(event, ServiceEventDeparture):
            await self.refresh_positions(event.vin)
        elif isinstance(event, ServiceEventOdometer):
            await self.refresh_info(event.vin)
            await self.refresh_maintenance(event.vin)

    async def _process_operation_event(self, event: OperationEvent) -> None:
        """Refresh the appropriate vehicle data based on the operation details."""
        _LOGGER.debug("Processing operation event: %s", event)
        if event.status == OperationStatus.ERROR:
            _LOGGER.warning(
                "Error received from car in operation %s, reason: %s.",
                event.status,
                event.error_code,
            )
            return
        if event.status == OperationStatus.IN_PROGRESS:
            return
        # The API backend doesn't seem to update right away after an operation completes so delay
        # a little bit before refreshing data. Magic numbers are bad but there is no way for us
        # to know when the backend has updated data...
        await asyncio.sleep(OPERATION_REFRESH_DELAY_SECONDS)
        if event.operation in [
            OperationName.STOP_AIR_CONDITIONING,
            OperationName.START_AIR_CONDITIONING,
            OperationName.SET_AIR_CONDITIONING_TARGET_TEMPERATURE,
            OperationName.START_WINDOW_HEATING,
            OperationName.STOP_WINDOW_HEATING,
            OperationName.SET_AIR_CONDITIONING_TIMERS,
        ]:
            await self.refresh_air_conditioning(event.vin)
        elif event.operation in [
            OperationName.START_AUXILIARY_HEATING,
            OperationName.STOP_AUXILIARY_HEATING,
        ]:
            await self.refresh_auxiliary_heating(event.vin)
        elif event.operation in [
            OperationName.UPDATE_CHARGE_LIMIT,
            OperationName.UPDATE_CARE_MODE,
            OperationName.UPDATE_CHARGING_CURRENT,
            OperationName.START_CHARGING,
            OperationName.STOP_CHARGING,
            OperationName.UPDATE_AUTO_UNLOCK_PLUG,
        ]:
            await self.refresh_charging(event.vin)
        elif event.operation in [
            OperationName.LOCK,
            OperationName.UNLOCK,
        ]:
            await self.refresh_status(event.vin)
        elif event.operation in [
            OperationName.UPDATE_DEPARTURE_TIMERS,
        ]:
            await self.refresh_departure_info(event.vin)

    async def _process_charging_event(self, event: ServiceEventChangeSoc) -> None:
        """Update self._vehicles with data from the event.

        Start by fully refreshing Vehicle.charging and Vehicle.driving_range as the endpoints
        may return updated data which is not included in the event. At the same time, the event
        may have more recent data so still apply data extracted from the event on top...
        """
        _LOGGER.debug("Processing charging event: %s", event)
        await self.refresh_charging(event.vin, notify=False)
        await self.refresh_driving_range(event.vin, notify=False)

        vehicle = self._vehicles[event.vin]
        if vehicle.charging and (status := vehicle.charging.status):
            self._process_charging_event_update_charging(status, event.data)

        if driving_range := vehicle.driving_range:
            self._process_charging_event_update_driving_range(driving_range, event.data)

        self._notify_callbacks(event.vin)

    @staticmethod
    def _process_charging_event_update_charging(
        charging_status: ChargingStatus, event_data: ServiceEventChangeSocData
    ) -> None:
        """Update charging_status with the event_data."""
        if event_data.charged_range:
            charging_status.battery.remaining_cruising_range_in_meters = (
                event_data.charged_range * 1000
            )
        if event_data.soc:
            charging_status.battery.state_of_charge_in_percent = event_data.soc
        if event_data.time_to_finish:
            charging_status.remaining_time_to_fully_charged_in_minutes = event_data.time_to_finish
        if event_data.state:
            charging_status.state = event_data.state

    @staticmethod
    def _process_charging_event_update_driving_range(
        driving_range: DrivingRange, event_data: ServiceEventChangeSocData
    ) -> None:
        """Update driving_range with the event_data."""
        per = driving_range.primary_engine_range
        ser = False

        if driving_range.secondary_engine_range:
            ser = driving_range.secondary_engine_range

        if event_data.soc:
            if per.engine_type == EngineType.ELECTRIC:
                per.current_soc_in_percent = event_data.soc
            elif ser and ser.engine_type == EngineType.ELECTRIC:
                ser.current_soc_in_percent = event_data.soc

        if event_data.charged_range:
            range_in_km = int(event_data.charged_range / 1000)
            if per.engine_type == EngineType.ELECTRIC:
                per.remaining_range_in_km = range_in_km
            elif ser and ser.engine_type == EngineType.ELECTRIC:
                ser.remaining_range_in_km = range_in_km



"""Models for responses of api/v2/garage/vehicles/{vin}."""

import logging
from dataclasses import dataclass, field
from enum import StrEnum

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin

from .common import BaseResponse
from .info import CompositeRender, Render, VehicleState

_LOGGER = logging.getLogger(__name__)


class GarageErrorType(StrEnum):
    """Known errors in the Garage."""

    NO_MOD1_4_VEHICLES = "NO_MOD1_4_VEHICLES"
    UNKNOWN = "UNKNOWN"
    MISSING_RENDER = "MISSING_RENDER"


@dataclass
class GarageError(DataClassORJSONMixin):
    """Errors occurring in the Garage."""

    description: str
    type: GarageErrorType


@dataclass
class GarageEntry(DataClassORJSONMixin):
    """One vehicle in the list of vehicles."""

    vin: str
    name: str
    state: VehicleState
    title: str
    priority: int
    device_platform: str = field(metadata=field_options(alias="devicePlatform"))
    system_model_id: str = field(metadata=field_options(alias="systemModelId"))
    renders: list[Render]
    composite_renders: list[CompositeRender] = field(
        metadata=field_options(alias="compositeRenders")
    )


@dataclass
class Garage(BaseResponse):
    """Contents of the users Garage."""

    vehicles: list[GarageEntry] | None = field(default=None)
    errors: list[GarageError] | None = field(default=None)



"""Models for MQTT Events."""

from .base import BaseEvent, EventType
from .operation import OperationEvent, OperationName, OperationStatus
from .service import (
    ServiceEvent,
    ServiceEventAccess,
    ServiceEventAirConditioning,
    ServiceEventChangeAccess,
    ServiceEventChangeLights,
    ServiceEventChangeOdometer,
    ServiceEventChangeSoc,
    ServiceEventChangeSocData,
    ServiceEventCharging,
    ServiceEventChargingCompleted,
    ServiceEventChargingError,
    ServiceEventData,
    ServiceEventDeparture,
    ServiceEventDepartureErrorPlug,
    ServiceEventError,
    ServiceEventErrorData,
    ServiceEventName,
    ServiceEventOdometer,
)
from .vehicle import (
    VehicleEventAwake,
    VehicleEventConnectionOnline,
    VehicleEventData,
    VehicleEventIgnitionStatusChanged,
    VehicleEventName,
    VehicleEventVehicleIgnitionStatusData,
    VehicleEventWarningBatterylevel,
)

__all__ = [
    "BaseEvent",
    "EventType",
    "OperationEvent",
    "OperationName",
    "OperationStatus",
    "ServiceEvent",
    "ServiceEventAccess",
    "ServiceEventAirConditioning",
    "ServiceEventChangeAccess",
    "ServiceEventChangeLights",
    "ServiceEventChangeOdometer",
    "ServiceEventChangeSoc",
    "ServiceEventChangeSocData",
    "ServiceEventCharging",
    "ServiceEventChargingCompleted",
    "ServiceEventChargingError",
    "ServiceEventData",
    "ServiceEventDeparture",
    "ServiceEventDepartureErrorPlug",
    "ServiceEventError",
    "ServiceEventErrorData",
    "ServiceEventName",
    "ServiceEventOdometer",
    "VehicleEventAwake",
    "VehicleEventConnectionOnline",
    "VehicleEventData",
    "VehicleEventIgnitionStatusChanged",
    "VehicleEventName",
    "VehicleEventVehicleIgnitionStatusData",
    "VehicleEventWarningBatterylevel",
]



"""Models related to vehicle events from the MQTT broker.

Vehicle Event example:
{
    "version": 1,
    "traceId": "800a74737b5a4328862d958c35b71b74",
    "producer": "SKODA_MHUB",
    "name": "vehicle-awake",
    "timestamp": "2025-05-11T07:35:18Z",
    "data": {
        "userId": "b8bc126c-ee36-402b-8723-2c1c3dff8dec",
        "vin": "TMOCKAA0AA000000",
    },
}
"""

from dataclasses import dataclass, field
from enum import StrEnum

from mashumaro import field_options
from mashumaro.types import Discriminator

from myskoda.models.vehicle_ignition_status import (
    IgnitionStatus,
    UnexpectedIgnitionStatusError,
)

from .base import BaseEvent, BaseEventData, EventType


class VehicleEventTopic(StrEnum):
    VEHICLE_CONNECTION_STATUS_UPDATE = "VEHICLE_CONNECTION_STATUS_UPDATE"
    VEHICLE_IGNITION_STATUS = "VEHICLE_IGNITION_STATUS"


class VehicleEventName(StrEnum):
    """List of known vehicle EventNames."""

    VEHICLE_AWAKE = "vehicle-awake"
    VEHICLE_CONNECTION_ONLINE = "vehicle-connection-online"
    VEHICLE_WARNING_BATTEYLEVEL = "vehicle-warning-batterylevel"
    VEHICLE_IGNITION_STATUS_CHANGED = "vehicle-ignition-status-changed"


def _deserialize_ignition_status(value: str) -> IgnitionStatus:
    match value:
        case "ON":
            return IgnitionStatus.ON
        case "OFF":
            return IgnitionStatus.OFF
        case _:
            raise UnexpectedIgnitionStatusError


@dataclass(frozen=True)
class VehicleEventData(BaseEventData):
    """Base model for data in all Vehicle Events."""


@dataclass(frozen=True)
class VehicleEvent(BaseEvent):
    """Base model for all Vehicle Events.

    Vehicle Events are unsolicited events emitted by the MQTT bus towards the client.
    Vehicle Events have a 'name' field which can be used as a discriminator.
    """

    class Config(BaseEvent.Config):  # noqa: D106
        discriminator = Discriminator(field="name", include_subtypes=True)

    event_type = EventType.VEHICLE_EVENT
    producer: str
    name: VehicleEventName
    data: VehicleEventData


@dataclass(frozen=True)
class VehicleEventVehicleIgnitionStatusData(VehicleEventData):
    """Ignition data inside vehicle service event vehicle-ignition-status."""

    ignition_status: IgnitionStatus | None = field(
        default=None,
        metadata=field_options(alias="ignitionStatus", deserialize=_deserialize_ignition_status),
    )


@dataclass(frozen=True)
class VehicleEventAwake(VehicleEvent):
    name = VehicleEventName.VEHICLE_AWAKE


@dataclass(frozen=True)
class VehicleEventConnectionOnline(VehicleEvent):
    name = VehicleEventName.VEHICLE_CONNECTION_ONLINE


@dataclass(frozen=True)
class VehicleEventIgnitionStatusChanged(VehicleEvent):
    name = VehicleEventName.VEHICLE_IGNITION_STATUS_CHANGED
    data: VehicleEventVehicleIgnitionStatusData


@dataclass(frozen=True)
class VehicleEventWarningBatterylevel(VehicleEvent):
    name = VehicleEventName.VEHICLE_WARNING_BATTEYLEVEL



"""Unit tests for MQTT."""

import asyncio
import json
from datetime import datetime
from pathlib import Path
from unittest.mock import ANY

import aiomqtt
import pytest

from myskoda.anonymize import USER_ID, VIN
from myskoda.models.charging import ChargeMode, ChargingState
from myskoda.models.event import (
    BaseEvent,
    EventType,
    OperationEvent,
    OperationName,
    OperationStatus,
    ServiceEventChangeAccess,
    ServiceEventChangeLights,
    ServiceEventChangeOdometer,
    ServiceEventChangeSoc,
    ServiceEventChangeSocData,
    ServiceEventChargingCompleted,
    ServiceEventData,
    ServiceEventName,
    VehicleEventAwake,
    VehicleEventConnectionOnline,
    VehicleEventData,
    VehicleEventIgnitionStatusChanged,
    VehicleEventName,
    VehicleEventVehicleIgnitionStatusData,
    VehicleEventWarningBatterylevel,
)
from myskoda.models.vehicle_ignition_status import IgnitionStatus
from myskoda.mqtt import MySkodaMqttClient

from .conftest import FakeMqttClientWrapper

FIXTURES_DIR = Path(__file__).parent / "fixtures"


@pytest.fixture
async def connected_mqtt_client(myskoda_mqtt_client: MySkodaMqttClient) -> MySkodaMqttClient:
    """Return a connected MySkodaMqttClient instance to use in tests."""
    await myskoda_mqtt_client.connect(user_id="1234", vehicle_vins=["TMOCKAA0AA000000"])
    return myskoda_mqtt_client


@pytest.mark.asyncio
async def test_wait_for_operation(
    connected_mqtt_client: MySkodaMqttClient, fake_mqtt_client_wrapper: FakeMqttClientWrapper
) -> None:
    topic = f"{USER_ID}/{VIN}/operation-request/air-conditioning/start-stop-air-conditioning"
    future = connected_mqtt_client.wait_for_operation(OperationName.START_AIR_CONDITIONING)

    assert future.done() is False

    in_progress = json.dumps(
        {
            "version": 1,
            "operation": "start-air-conditioning",
            "status": "IN_PROGRESS",
            "traceId": "f0b638f7bba08ec4acb5de64cdb97ba9",
            "requestId": "72f24950-b3db-4b7e-948f-7032f533773a",
        }
    )
    complete = json.dumps(
        {
            "version": 1,
            "operation": "start-air-conditioning",
            "status": "COMPLETED_SUCCESS",
            "traceId": "10d37beb6b37e9a9e74460d402855f27",
            "requestId": "72f24950-b3db-4b7e-948f-7032f533773a",
        }
    )
    fake_mqtt_client_wrapper.set_messages(
        [
            aiomqtt.Message(
                topic=topic, payload=in_progress, qos=1, retain=False, mid=1, properties=None
            ),
            aiomqtt.Message(
                topic=topic, payload=complete, qos=1, retain=False, mid=1, properties=None
            ),
        ]
    )

    await future


@pytest.mark.asyncio
async def test_subscribe_event(
    connected_mqtt_client: MySkodaMqttClient, fake_mqtt_client_wrapper: FakeMqttClientWrapper
) -> None:
    base_topic = f"{USER_ID}/{VIN}"
    trace_id = "7a59299d06535a6756d10e96e0c75ed3"
    request_id = "b9bc1258-2d0c-43c2-8d67-44d9f6c8cb9f"
    timestamp_str = "2024-10-17T08:49:59.538Z"
    timestamp = datetime.fromisoformat(timestamp_str)

    events: list[BaseEvent] = []
    future = asyncio.get_event_loop().create_future()

    messages = [
        aiomqtt.Message(
            topic=f"{base_topic}/operation-request/air-conditioning/start-stop-air-conditioning",
            payload=json.dumps(
                {
                    "version": 1,
                    "operation": "stop-air-conditioning",
                    "status": "COMPLETED_SUCCESS",
                    "traceId": trace_id,
                    "requestId": request_id,
                }
            ),
            qos=1,
            retain=False,
            mid=1,
            properties=None,
        ),
        aiomqtt.Message(
            topic=f"{base_topic}/service-event/vehicle-status/lights",
            payload=json.dumps(
                {
                    "version": 1,
                    "traceId": trace_id,
                    "timestamp": timestamp_str,
                    "producer": "SKODA_MHUB",
                    "name": "change-lights",
                    "data": {
                        "userId": USER_ID,
                        "vin": VIN,
                    },
                }
            ),
            qos=1,
            retain=False,
            mid=1,
            properties=None,
        ),
        aiomqtt.Message(
            topic=f"{base_topic}/service-event/vehicle-status/odometer",
            payload=json.dumps(
                {
                    "version": 1,
                    "traceId": trace_id,
                    "timestamp": timestamp_str,
                    "producer": "SKODA_MHUB",
                    "name": "change-odometer",
                    "data": {
                        "userId": USER_ID,
                        "vin": VIN,
                    },
                }
            ),
            qos=1,
            retain=False,
            mid=1,
            properties=None,
        ),
        aiomqtt.Message(
            topic=f"{base_topic}/service-event/vehicle-status/access",
            payload=json.dumps(
                {
                    "version": 1,
                    "traceId": trace_id,
                    "timestamp": timestamp_str,
                    "producer": "SKODA_MHUB",
                    "name": "change-access",
                    "data": {
                        "userId": USER_ID,
                        "vin": VIN,
                    },
                }
            ),
            qos=1,
            retain=False,
            mid=1,
            properties=None,
        ),
        aiomqtt.Message(
            topic=f"{base_topic}/service-event/charging",
            payload=json.dumps(
                {
                    "version": 1,
                    "traceId": trace_id,
                    "timestamp": timestamp_str,
                    "producer": "SKODA_MHUB",
                    "name": "change-soc",
                    "data": {
                        "mode": "manual",
                        "state": "charging",
                        "soc": "91",
                        "chargedRange": "307",
                        "timeToFinish": "40",
                        "userId": USER_ID,
                        "vin": VIN,
                    },
                }
            ),
            qos=1,
            retain=False,
            mid=1,
            properties=None,
        ),
        aiomqtt.Message(
            topic=f"{base_topic}/service-event/charging",
            payload=json.dumps(
                {
                    "version": 1,
                    "traceId": trace_id,
                    "timestamp": timestamp_str,
                    "producer": "SKODA_MHUB",
                    "name": "charging-completed",
                    "data": {
                        "mode": "manual",
                        "state": "chargePurposeReachedAndConservation",
                        "soc": "100",
                        "chargedRange": "500",
                        "timeToFinish": "0",
                        "userId": USER_ID,
                        "vin": VIN,
                    },
                }
            ),
            qos=1,
            retain=False,
            mid=1,
            properties=None,
        ),
        aiomqtt.Message(
            topic=f"{base_topic}/service-event/charging",
            payload=json.dumps(
                {
                    "version": 1,
                    "traceId": trace_id,
                    "timestamp": timestamp_str,
                    "producer": "SKODA_MHUB",
                    "name": "charging-completed",
                    "data": {
                        "userId": USER_ID,
                        "vin": VIN,
                    },
                }
            ),
            qos=1,
            retain=False,
            mid=1,
            properties=None,
        ),
        aiomqtt.Message(
            topic=f"{base_topic}/vehicle-event/vehicle-connection-status-update",
            payload=json.dumps(
                {
                    "version": 1,
                    "traceId": trace_id,
                    "timestamp": timestamp_str,
                    "producer": "SKODA_MHUB",
                    "name": "vehicle-connection-online",
                    "data": {
                        "userId": USER_ID,
                        "vin": VIN,
                    },
                }
            ),
            qos=1,
            retain=False,
            mid=1,
            properties=None,
        ),
        aiomqtt.Message(
            topic=f"{base_topic}/vehicle-event/vehicle-connection-status-update",
            payload=json.dumps(
                {
                    "version": 1,
                    "traceId": trace_id,
                    "timestamp": timestamp_str,
                    "producer": "SKODA_MHUB",
                    "name": "vehicle-awake",
                    "data": {
                        "userId": USER_ID,
                        "vin": VIN,
                    },
                }
            ),
            qos=1,
            retain=False,
            mid=1,
            properties=None,
        ),
        aiomqtt.Message(
            topic=f"{base_topic}/vehicle-event/vehicle-connection-status-update",
            payload=json.dumps(
                {
                    "version": 1,
                    "traceId": trace_id,
                    "timestamp": timestamp_str,
                    "producer": "SKODA_MHUB",
                    "name": "vehicle-warning-batterylevel",
                    "data": {
                        "userId": USER_ID,
                        "vin": VIN,
                    },
                }
            ),
            qos=1,
            retain=False,
            mid=1,
            properties=None,
        ),
        aiomqtt.Message(
            topic=f"{base_topic}/vehicle-event/vehicle-ignition-status",
            payload=json.dumps(
                {
                    "version": 1,
                    "traceId": trace_id,
                    "timestamp": timestamp_str,
                    "producer": "SKODA_MHUB",
                    "name": "vehicle-ignition-status-changed",
                    "data": {
                        "userId": USER_ID,
                        "vin": VIN,
                        "ignitionStatus": IgnitionStatus.ON,
                    },
                }
            ),
            qos=1,
            retain=False,
            mid=1,
            properties=None,
        ),
        aiomqtt.Message(
            topic=f"{base_topic}/vehicle-event/vehicle-ignition-status",
            payload=json.dumps(
                {
                    "version": 1,
                    "traceId": trace_id,
                    "timestamp": timestamp_str,
                    "producer": "SKODA_MHUB",
                    "name": "vehicle-ignition-status-changed",
                    "data": {
                        "userId": USER_ID,
                        "vin": VIN,
                        "ignitionStatus": IgnitionStatus.OFF,
                    },
                }
            ),
            qos=1,
            retain=False,
            mid=1,
            properties=None,
        ),
    ]

    async def on_event(event: BaseEvent) -> None:
        events.append(event)
        if len(events) == len(messages):
            future.set_result(None)

    connected_mqtt_client.subscribe(on_event)

    # @dvx76: not sure why messages get received/send in reverse order ...
    fake_mqtt_client_wrapper.set_messages(list(reversed(messages)))

    await future

    assert events == [
        OperationEvent(
            vin=VIN,
            event_type=EventType.OPERATION,
            timestamp=ANY,
            version=1,
            operation=OperationName.STOP_AIR_CONDITIONING,
            trace_id=trace_id,
            request_id=request_id,
            status=OperationStatus.COMPLETED_SUCCESS,
            error_code=None,
        ),
        ServiceEventChangeLights(
            vin=VIN,
            event_type=EventType.SERVICE_EVENT,
            version=1,
            trace_id=trace_id,
            timestamp=timestamp,
            producer="SKODA_MHUB",
            name=ServiceEventName.CHANGE_LIGHTS,
            data=ServiceEventData(user_id=USER_ID, vin=VIN),
        ),
        ServiceEventChangeOdometer(
            vin=VIN,
            event_type=EventType.SERVICE_EVENT,
            version=1,
            trace_id=trace_id,
            timestamp=timestamp,
            producer="SKODA_MHUB",
            name=ServiceEventName.CHANGE_ODOMETER,
            data=ServiceEventData(user_id=USER_ID, vin=VIN),
        ),
        ServiceEventChangeAccess(
            vin=VIN,
            event_type=EventType.SERVICE_EVENT,
            version=1,
            trace_id=trace_id,
            timestamp=timestamp,
            producer="SKODA_MHUB",
            name=ServiceEventName.CHANGE_ACCESS,
            data=ServiceEventData(user_id=USER_ID, vin=VIN),
        ),
        ServiceEventChangeSoc(
            vin=VIN,
            event_type=EventType.SERVICE_EVENT,
            version=1,
            trace_id=trace_id,
            timestamp=timestamp,
            producer="SKODA_MHUB",
            name=ServiceEventName.CHANGE_SOC,
            data=ServiceEventChangeSocData(
                user_id=USER_ID,
                vin=VIN,
                charged_range=307,
                soc=91,
                state=ChargingState.CHARGING,
                mode=ChargeMode.MANUAL,
                time_to_finish=40,
            ),
        ),
        ServiceEventChargingCompleted(
            vin=VIN,
            event_type=EventType.SERVICE_EVENT,
            version=1,
            trace_id=trace_id,
            timestamp=timestamp,
            producer="SKODA_MHUB",
            name=ServiceEventName.CHARGING_COMPLETED,
            data=ServiceEventChangeSocData(
                user_id=USER_ID,
                vin=VIN,
                charged_range=500,
                soc=100,
                state=ChargingState.CONSERVING,
                mode=ChargeMode.MANUAL,
                time_to_finish=0,
            ),
        ),
        ServiceEventChargingCompleted(
            vin=VIN,
            event_type=EventType.SERVICE_EVENT,
            version=1,
            trace_id=trace_id,
            timestamp=timestamp,
            producer="SKODA_MHUB",
            name=ServiceEventName.CHARGING_COMPLETED,
            data=ServiceEventChangeSocData(
                user_id=USER_ID,
                vin=VIN,
                mode=None,
                state=None,
                soc=None,
                charged_range=None,
                time_to_finish=None,
            ),
        ),
        VehicleEventConnectionOnline(
            vin=VIN,
            event_type=EventType.VEHICLE_EVENT,
            version=1,
            trace_id=trace_id,
            timestamp=timestamp,
            producer="SKODA_MHUB",
            name=VehicleEventName.VEHICLE_CONNECTION_ONLINE,
            data=VehicleEventData(user_id=USER_ID, vin=VIN),
        ),
        VehicleEventAwake(
            vin=VIN,
            event_type=EventType.VEHICLE_EVENT,
            version=1,
            trace_id=trace_id,
            timestamp=timestamp,
            producer="SKODA_MHUB",
            name=VehicleEventName.VEHICLE_AWAKE,
            data=VehicleEventData(user_id=USER_ID, vin=VIN),
        ),
        VehicleEventWarningBatterylevel(
            vin=VIN,
            event_type=EventType.VEHICLE_EVENT,
            version=1,
            trace_id=trace_id,
            timestamp=timestamp,
            producer="SKODA_MHUB",
            name=VehicleEventName.VEHICLE_WARNING_BATTEYLEVEL,
            data=VehicleEventData(user_id=USER_ID, vin=VIN),
        ),
        VehicleEventIgnitionStatusChanged(
            vin=VIN,
            event_type=EventType.VEHICLE_EVENT,
            version=1,
            trace_id=trace_id,
            timestamp=timestamp,
            producer="SKODA_MHUB",
            name=VehicleEventName.VEHICLE_IGNITION_STATUS_CHANGED,
            data=VehicleEventVehicleIgnitionStatusData(
                user_id=USER_ID, vin=VIN, ignition_status=IgnitionStatus.ON
            ),
        ),
        VehicleEventIgnitionStatusChanged(
            vin=VIN,
            event_type=EventType.VEHICLE_EVENT,
            version=1,
            trace_id=trace_id,
            timestamp=timestamp,
            producer="SKODA_MHUB",
            name=VehicleEventName.VEHICLE_IGNITION_STATUS_CHANGED,
            data=VehicleEventVehicleIgnitionStatusData(
                user_id=USER_ID, vin=VIN, ignition_status=IgnitionStatus.OFF
            ),
        ),
    ]



"""Models for responses of api/v2/garage/vehicles/{vin}."""

import logging
from dataclasses import dataclass, field
from datetime import date
from enum import StrEnum

from mashumaro import field_options
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.mixins.yaml import DataClassYAMLMixin

from .common import BaseResponse

_LOGGER = logging.getLogger(__name__)


class CapabilityId(StrEnum):
    """List of known Capabilities."""

    ACCESS = "ACCESS"
    ACCIDENT_DAMAGE_MANAGEMENT = "ACCIDENT_DAMAGE_MANAGEMENT"
    ACTIVE_VENTILATION = "ACTIVE_VENTILATION"
    AIR_CONDITIONING = "AIR_CONDITIONING"
    AIR_CONDITIONING_HEATING_SOURCE_AUXILIARY = "AIR_CONDITIONING_HEATING_SOURCE_AUXILIARY"
    AIR_CONDITIONING_HEATING_SOURCE_ELECTRIC = "AIR_CONDITIONING_HEATING_SOURCE_ELECTRIC"
    AIR_CONDITIONING_SAVE_AND_ACTIVATE = "AIR_CONDITIONING_SAVE_AND_ACTIVATE"
    AIR_CONDITIONING_SMART_SETTINGS = "AIR_CONDITIONING_SMART_SETTINGS"
    AIR_CONDITIONING_TIMERS = "AIR_CONDITIONING_TIMERS"
    AUTOMATION = "AUTOMATION"
    AUXILIARY_HEATING = "AUXILIARY_HEATING"
    AUXILIARY_HEATING_BASIC = "AUXILIARY_HEATING_BASIC"
    AUXILIARY_HEATING_TEMPERATURE_SETTING = "AUXILIARY_HEATING_TEMPERATURE_SETTING"
    AUXILIARY_HEATING_TIMERS = "AUXILIARY_HEATING_TIMERS"
    BATTERY_CHARGING_CARE = "BATTERY_CHARGING_CARE"
    BATTERY_SUPPORT = "BATTERY_SUPPORT"
    CAR_FEEDBACK = "CAR_FEEDBACK"
    CARE_AND_INSURANCE = "CARE_AND_INSURANCE"
    CHARGE_MODE_SELECTION = "CHARGE_MODE_SELECTION"
    CHARGING = "CHARGING"
    CHARGING_MEB = "CHARGING_MEB"
    CHARGING_MQB = "CHARGING_MQB"
    CHARGING_PROFILES = "CHARGING_PROFILES"
    CHARGING_STATIONS = "CHARGING_STATIONS"
    CUBIC = "CUBIC"
    DEALER_APPOINTMENT = "DEALER_APPOINTMENT"
    DEPARTURE_TIMERS = "DEPARTURE_TIMERS"
    DESTINATIONS = "DESTINATIONS"
    DESTINATION_IMPORT = "DESTINATION_IMPORT"
    DESTINATION_IMPORT_UPGRADABLE = "DESTINATION_IMPORT_UPGRADABLE"
    DIGICERT = "DIGICERT"
    DRIVING_SCORE = "DRIVING_SCORE"
    EMERGENCY_CALLING = "EMERGENCY_CALLING"
    EV_ROUTE_PLANNING = "EV_ROUTE_PLANNING"
    EXTENDED_CHARGING_SETTINGS = "EXTENDED_CHARGING_SETTINGS"
    FLEET_SUPPORTED = "FLEET_SUPPORTED"
    FUEL_STATUS = "FUEL_STATUS"
    GEO_FENCE = "GEO_FENCE"
    GUEST_USER_MANAGEMENT = "GUEST_USER_MANAGEMENT"
    HONK_AND_FLASH = "HONK_AND_FLASH"
    ICE_VEHICLE_RTS = "ICE_VEHICLE_RTS"
    LOYALTY_PROGRAM = "LOYALTY_PROGRAM"
    LOYALTY_PROGRAM_WORLDWIDE = "LOYALTY_PROGRAM_WORLDWIDE"
    MAP_UPDATE = "MAP_UPDATE"
    MEASUREMENTS = "MEASUREMENTS"
    MISUSE_PROTECTION = "MISUSE_PROTECTION"
    NEWS = "NEWS"
    ONLINE_REMOTE_UPDATE = "ONLINE_REMOTE_UPDATE"
    ONLINE_SPEECH_GPS = "ONLINE_SPEECH_GPS"
    OUTSIDE_TEMPERATURE = "OUTSIDE_TEMPERATURE"
    PARKING_INFORMATION = "PARKING_INFORMATION"
    PARKING_POSITION = "PARKING_POSITION"
    PAY_TO_FUEL = "PAY_TO_FUEL"
    PAY_TO_PARK = "PAY_TO_PARK"
    PLUG_AND_CHARGE = "PLUG_AND_CHARGE"
    POI_SEARCH = "POI_SEARCH"
    POWERPASS_TARIFFS = "POWERPASS_TARIFFS"
    PREDICTIVE_WAKE_UP = "PREDICTIVE_WAKE_UP"
    READINESS = "READINESS"
    ROADSIDE_ASSISTANT = "ROADSIDE_ASSISTANT"
    ROUTE_IMPORT = "ROUTE_IMPORT"
    ROUTE_PLANNING_5_CHARGERS = "ROUTE_PLANNING_5_CHARGERS"
    ROUTE_PLANNING_10_CHARGERS = "ROUTE_PLANNING_10_CHARGERS"
    ROUTING = "ROUTING"
    SERVICE_PARTNER = "SERVICE_PARTNER"
    SPEED_ALERT = "SPEED_ALERT"
    STATE = "STATE"
    SUBSCRIPTIONS = "SUBSCRIPTIONS"
    THEFT_WARNING = "THEFT_WARNING"
    TRAFFIC_INFORMATION = "TRAFFIC_INFORMATION"
    TRIP_STATISTICS = "TRIP_STATISTICS"
    UNAVAILABILITY_STATUSES = "UNAVAILABILITY_STATUSES"
    VEHICLE_HEALTH_INSPECTION = "VEHICLE_HEALTH_INSPECTION"
    VEHICLE_HEALTH_WARNINGS = "VEHICLE_HEALTH_WARNINGS"
    VEHICLE_HEALTH_WARNINGS_WITH_WAKE_UP = "VEHICLE_HEALTH_WARNINGS_WITH_WAKE_UP"
    VEHICLE_SERVICES_BACKUPS = "VEHICLE_SERVICES_BACKUPS"
    VEHICLE_WAKE_UP = "VEHICLE_WAKE_UP"
    VEHICLE_WAKE_UP_TRIGGER = "VEHICLE_WAKE_UP_TRIGGER"
    WARNING_LIGHTS = "WARNING_LIGHTS"
    WEB_RADIO = "WEB_RADIO"
    WINDOW_HEATING = "WINDOW_HEATING"


class CapabilityStatus(StrEnum):
    """List of known statuses for Capabilities."""

    DEACTIVATED_BY_ACTIVE_VEHICLE_USER = "DEACTIVATED_BY_ACTIVE_VEHICLE_USER"
    DISABLED_BY_USER = "DISABLED_BY_USER"
    FRONTEND_SWITCHED_OFF = "FRONTEND_SWITCHED_OFF"
    INITIALLY_DISABLED = "INITIALLY_DISABLED"
    INSUFFICIENT_BATTERY_LEVEL = "INSUFFICIENT_BATTERY_LEVEL"
    INSUFFICIENT_RIGHTS = "INSUFFICIENT_RIGHTS"
    INSUFFICIENT_SPIN = "INSUFFICIENT_SPIN"
    LICENSE_EXPIRED = "LICENSE_EXPIRED"
    LICENSE_MISSING = "LICENSE_MISSING"
    LOCATION_DATA_DISABLED = "LOCATION_DATA_DISABLED"


class ErrorType(StrEnum):
    """Known errors."""

    MISSING_RENDER = "MISSING_RENDER"
    UNAVAILABLE_SERVICE_PLATFORM_CAPABILITIES = "UNAVAILABLE_SERVICE_PLATFORM_CAPABILITIES"
    UNAVAILABLE_SOFTWARE_VERSION = "UNAVAILABLE_SOFTWARE_VERSION"


@dataclass
class Error(DataClassORJSONMixin):
    """Main model for emitted errors."""

    description: str
    type: ErrorType


@dataclass
class Capability(DataClassORJSONMixin, DataClassYAMLMixin):
    """Shows the status of a capability. Empty status indicates no error."""

    id: CapabilityId
    statuses: list[CapabilityStatus]

    def is_available(self) -> bool:
        """Check whether the capability can currently be used.

        It looks like every status is an indication that the capability is not available.
        """
        return not self.statuses


def drop_unknown_capabilities(value: list[dict]) -> list[Capability]:
    """Drop any unknown capabilities and log a message."""
    unknown_capabilities = [c for c in value if c["id"] not in CapabilityId]
    if unknown_capabilities:
        _LOGGER.info("Dropping unknown capabilities: %s", unknown_capabilities)
    return [Capability.from_dict(c) for c in value if c["id"] in CapabilityId]


@dataclass
class Capabilities(DataClassORJSONMixin):
    """Main Model for Capabilities.

    Capabilities are Skoda software features known by the library.
    """

    capabilities: list[Capability] = field(
        metadata=field_options(deserialize=drop_unknown_capabilities)
    )
    errors: list[Error] | None = field(default=None)


@dataclass
class Battery(DataClassORJSONMixin):
    """Battery features."""

    capacity: int = field(metadata=field_options(alias="capacityInKWh"))


class BodyType(StrEnum):
    """Known car body types."""

    SUV = "SUV"
    SUV_COUPE = "SUV Coupe"
    COMBI = "Combi"
    LIFTBACK = "Liftback"
    HATCHBACK = "Hatchback"
    CROSSOVER = "Crossover"
    SPACEBACK = "Spaceback"


class VehicleState(StrEnum):
    """Main software state of the vehicle."""

    ACTIVATED = "ACTIVATED"
    GUEST_USER = "GUEST_USER"
    NOT_ACTIVATED = "NOT_ACTIVATED"
    RESET_SPIN = "RESET_SPIN"


@dataclass
class Engine(DataClassORJSONMixin):
    """Engine features."""

    power: int = field(metadata=field_options(alias="powerInKW"))
    type: str | None = field(default=None)
    capacity_in_liters: float | None = field(
        default=None, metadata=field_options(alias="capacityInLiters")
    )


@dataclass
class Gearbox(DataClassORJSONMixin):
    """Gearbox features."""

    type: str


@dataclass
class Specification(DataClassORJSONMixin):
    """Car specification. Model for the physical features of the car."""

    body: BodyType
    engine: Engine
    model: str
    title: str
    manufacturing_date: date = field(metadata=field_options(alias="manufacturingDate"))
    model_year: str = field(metadata=field_options(alias="modelYear"))
    system_code: str = field(metadata=field_options(alias="systemCode"))
    system_model_id: str = field(metadata=field_options(alias="systemModelId"))
    battery: Battery | None = field(default=None)
    max_charging_power: int | None = field(
        default=None, metadata=field_options(alias="maxChargingPowerInKW")
    )
    trim_level: str | None = field(default=None, metadata=field_options(alias="trimLevel"))


@dataclass
class ServicePartner(DataClassORJSONMixin):
    """ServicePartner is a fancy name for car dealer."""

    id: str = field(metadata=field_options(alias="servicePartnerId"))


class ViewType(StrEnum):
    UNMODIFIED_EXTERIOR_SIDE = "UNMODIFIED_EXTERIOR_SIDE"
    UNMODIFIED_EXTERIOR_FRONT = "UNMODIFIED_EXTERIOR_FRONT"
    HOME = "HOME"
    CHARGING_LIGHT = "CHARGING_LIGHT"
    CHARGING_DARK = "CHARGING_DARK"
    PLUGGED_IN_DARK = "PLUGGED_IN_DARK"
    PLUGGED_IN_LIGHT = "PLUGGED_IN_LIGHT"


class RenderType(StrEnum):
    REAL = "REAL"


@dataclass
class Render(DataClassORJSONMixin):
    url: str
    type: RenderType
    order: int
    view_point: str = field(metadata=field_options(alias="viewPoint"))


@dataclass
class CompositeRender(DataClassORJSONMixin):
    layers: list[Render]
    view_type: ViewType = field(metadata=field_options(alias="viewType"))


@dataclass
class Info(BaseResponse):
    """Basic vehicle information."""

    state: VehicleState
    specification: Specification
    vin: str
    name: str
    capabilities: Capabilities
    renders: list[Render]
    device_platform: str = field(metadata=field_options(alias="devicePlatform"))
    workshop_mode_enabled: bool = field(metadata=field_options(alias="workshopModeEnabled"))
    composite_renders: list[CompositeRender] = field(
        metadata=field_options(alias="compositeRenders")
    )
    service_partner: ServicePartner | None = field(
        default=None, metadata=field_options(alias="servicePartner")
    )
    software_version: str | None = field(
        default=None, metadata=field_options(alias="softwareVersion")
    )
    license_plate: str | None = field(default=None, metadata=field_options(alias="licensePlate"))
    errors: list[Error] | None = field(default=None)

    def has_capability(self, cap: CapabilityId) -> bool:
        """Check for a capability.

        Checks whether a vehicle generally has a capability.
        Does not check whether it's actually available.
        """
        return any(capability.id == cap for capability in self.capabilities.capabilities)

    def is_capability_available(self, cap: CapabilityId) -> bool:
        """Check for capability availability.

        Checks whether the vehicle has the capability and whether it is currently
        available. A capability can be unavailable for example if it's deactivated
        by the currently active user.
        """
        return any(
            capability.id == cap and capability.is_available()
            for capability in self.capabilities.capabilities
        )

    def get_model_name(self) -> str:
        """Return the name of the vehicle's model."""
        model = self.specification.model
        engine = self.specification.engine
        model_year = self.specification.model_year
        system_model_id = self.specification.system_model_id
        return f"{model} {engine} {model_year} ({system_model_id})"



"""Events emitted by MySkoda."""

import json
from dataclasses import dataclass, field
from datetime import UTC, datetime
from enum import StrEnum

from mashumaro import field_options
from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import Discriminator


class EventType(StrEnum):
    """The different 'types' of events.

    The type is determine by a part of the MQTT topic.
    """

    ACCOUNT_EVENT = "account-event"
    OPERATION = "operation-request"
    SERVICE_EVENT = "service-event"
    VEHICLE_EVENT = "vehicle-event"
    UNKNOWN = "unknown"


@dataclass(frozen=True)
class BaseEventData(DataClassORJSONMixin):
    """Base model for data in all Events."""

    user_id: str = field(metadata=field_options(alias="userId"))
    vin: str


@dataclass(frozen=True, kw_only=True)
class BaseEvent(DataClassORJSONMixin):
    """Base model for all Events.

    Raw events from the MQTT broker don't include vin or event_type.
    """

    class Config(BaseConfig):  # noqa: D106
        discriminator = Discriminator(field="event_type", include_subtypes=True)

    vin: str = ""
    event_type: EventType = EventType.UNKNOWN
    version: int
    trace_id: str = field(metadata=field_options(alias="traceId"))
    timestamp: datetime = field(default_factory=lambda: datetime.now(UTC))

    @classmethod
    def from_mqtt_message(cls, topic: str, payload: str) -> "BaseEvent":
        """Return a parsed event object.

        'topic' is the original MQTT topic on which the event was received
        'payload' is the original payload as a binary string
        """
        # Depending on the event the event_topic can be one or two levels of the mqtt topic
        _user_id, vin, event_type, _event_topic = topic.split("/", maxsplit=3)
        deserialized_payload = json.loads(payload)
        deserialized_payload["vin"] = vin
        deserialized_payload["event_type"] = event_type
        return cls.from_dict(deserialized_payload)



"""MQTT client module for the MySkoda server.

Inspired by https://github.com/YoSmart-Inc/yolink-api/tree/main
"""

import asyncio
import logging
import re
import ssl
import uuid
from collections.abc import AsyncIterator, Callable, Coroutine
from random import uniform
from types import TracebackType
from typing import Any, Protocol, Self, cast

import aiomqtt

from .auth.authorization import Authorization
from .const import (
    MQTT_ACCOUNT_EVENT_TOPICS,
    MQTT_BROKER_HOST,
    MQTT_BROKER_PORT,
    MQTT_FAST_RETRY,
    MQTT_KEEPALIVE,
    MQTT_MAX_RECONNECT_DELAY,
    MQTT_OPERATION_TOPICS,
    MQTT_RECONNECT_DELAY,
    MQTT_SERVICE_EVENT_TOPICS,
    MQTT_VEHICLE_EVENT_TOPICS,
)
from .models.event import BaseEvent, OperationEvent, OperationName, OperationStatus

_LOGGER = logging.getLogger(__name__)
TOPIC_RE = re.compile("^(.*?)/(.*?)/(.*?)/(.*?)$")
APP_UUID = uuid.uuid4()


def _create_ssl_context() -> ssl.SSLContext:
    """Create a SSL context for the MQTT connection."""
    context = ssl.SSLContext(ssl.PROTOCOL_TLS)
    context.load_default_certs()
    return context


_SSL_CONTEXT = _create_ssl_context()

background_tasks = set()


class OperationListener:
    """Used to track callbacks to execute for a given OperationName."""

    operation_name: OperationName
    future: asyncio.Future[OperationEvent]

    def __init__(
        self, operation_name: OperationName, future: asyncio.Future[OperationEvent]
    ) -> None:
        self.operation_name = operation_name
        self.future = future


class OperationFailedError(Exception):
    def __init__(self, event: OperationEvent) -> None:
        op = event.operation
        error = event.error_code
        trace = event.trace_id
        super().__init__(f"Operation {op} with trace {trace} failed: {error}")


class AbstractMqttClientWrapper(Protocol):
    """Interface for an aiomqtt.Client wrapper.

    We're using an interface so we can pass in a fake wrapper in tests.
    We're wrapping (technically subclassing) so we can add a clean public update_username_password
    method.
    """

    async def __aenter__(self) -> Self: ...  # noqa: D105

    async def __aexit__(  # noqa: D105
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None: ...

    @property
    def messages(self) -> AsyncIterator: ...  # noqa: D102

    async def subscribe(self, topic: str) -> None: ...  # noqa: D102

    def update_username_password(self, username: str, password: str) -> None: ...  # noqa: D102


class AioMqttClientWrapper(aiomqtt.Client):
    def update_username_password(self, username: str, password: str) -> None:
        """Update the username and password set in an aiomqtt.Client object.

        This isn't super clean since _client is a private member but it's just an instance of
        paho.mqtt.Client so unlikely to change.

        We want the ability to update the username/pw so that MySkodaMqttClient can instantiate
        this in its __init__ but defer setting the username/pw.
        """
        self._client.username_pw_set(username=username, password=password)


class MySkodaMqttClient:
    user_id: str | None
    vehicle_vins: list[str]
    _callbacks: list[Callable[[BaseEvent], Coroutine[Any, Any, None]]]
    _operation_listeners: list[OperationListener]

    def __init__(
        self,
        authorization: Authorization,
        mqtt_client: AbstractMqttClientWrapper | None = None,
        ssl_context: ssl.SSLContext | None = None,
    ) -> None:
        self.authorization = authorization
        self.vehicle_vins = []
        self.mqtt_client = mqtt_client
        if self.mqtt_client is None:
            # Pass in pre-created SSLContext (vs 'tls_params=aiomqtt.TLSParameters()') to avoid a
            # blocking call in paho.mqtt.client. See https://github.com/w1ll1am23/pyeconet/pull/43.
            self.mqtt_client = AioMqttClientWrapper(
                hostname=MQTT_BROKER_HOST,
                port=MQTT_BROKER_PORT,
                identifier="Id" + str(APP_UUID) + "#" + str(uuid.uuid4()),
                logger=_LOGGER,
                tls_context=ssl_context if ssl_context else _SSL_CONTEXT,
                keepalive=MQTT_KEEPALIVE,
                clean_session=True,
            )
        self._callbacks = []
        self._operation_listeners = []
        self._listener_task = None
        self._running = False
        self._subscribed = asyncio.Event()
        self._reconnect_delay = MQTT_RECONNECT_DELAY

    async def connect(self, user_id: str, vehicle_vins: list[str]) -> None:
        """Connect to the MQTT broker and listen for messages for the given user_id and VINs."""
        _LOGGER.info("Connecting to MQTT with %s/%s", user_id, vehicle_vins)
        self.user_id = user_id
        self.vehicle_vins = vehicle_vins
        self._listener_task = asyncio.create_task(self._connect_and_listen())
        await self._subscribed.wait()

    async def disconnect(self) -> None:
        """Cancel listener task and set self_running to False, causing the listen loop to end."""
        if self._listener_task is None:
            return
        self._listener_task.cancel()
        self._listener_task = None
        self._running = False

    def subscribe(self, callback: Callable[[BaseEvent], Coroutine[Any, Any, None]]) -> None:
        """Listen for events emitted by MySkoda's MQTT broker."""
        self._callbacks.append(callback)

    def wait_for_operation(self, operation_name: OperationName) -> asyncio.Future[OperationEvent]:
        """Wait until the next operation of the specified type completes."""
        _LOGGER.debug("Waiting for operation %s complete.", operation_name)
        future: asyncio.Future[OperationEvent] = asyncio.get_event_loop().create_future()

        self._operation_listeners.append(OperationListener(operation_name, future))

        return future

    async def _connect_and_listen(self) -> None:
        """Connect to the MQTT broker and listen for messages for the given user_id and VINs.

        Reconnect loop based on https://github.com/empicano/aiomqtt/blob/main/docs/reconnection.md.

        Re-fetch and update the authorization token on every try.
        """
        _LOGGER.debug("Starting _connect_and_listen")
        self._running = True
        retry_count = 0  # Track the number of retries
        self._reconnect_delay = MQTT_RECONNECT_DELAY  # Initial delay for backoff
        while self._running:
            try:
                assert self.mqtt_client is not None
                password = await self.authorization.get_access_token()
                self.mqtt_client.update_username_password(username="android-app", password=password)
                async with self.mqtt_client as client:
                    _LOGGER.info("Connected to MQTT")
                    _LOGGER.debug("using MQTT client %s", client)
                    for vin in self.vehicle_vins:
                        for topic in MQTT_OPERATION_TOPICS:
                            await client.subscribe(
                                f"{self.user_id}/{vin}/operation-request/{topic}"
                            )
                        for topic in MQTT_SERVICE_EVENT_TOPICS:
                            await client.subscribe(f"{self.user_id}/{vin}/service-event/{topic}")
                        for topic in MQTT_ACCOUNT_EVENT_TOPICS:
                            await client.subscribe(f"{self.user_id}/{vin}/account-event/{topic}")
                        for topic in MQTT_VEHICLE_EVENT_TOPICS:
                            await client.subscribe(f"{self.user_id}/{vin}/vehicle-event/{topic}")

                    self._subscribed.set()
                    self._reconnect_delay = MQTT_RECONNECT_DELAY
                    retry_count = 0  # Reset retry count on successful connection
                    async for message in client.messages:
                        self._on_message(message)
            except aiomqtt.MqttError as exc:
                retry_count += 1
                _LOGGER.info(
                    "Connection lost (%s); reconnecting in %ss", exc, self._reconnect_delay
                )
                await asyncio.sleep(self._reconnect_delay)
                if (
                    retry_count > MQTT_FAST_RETRY
                    and self._reconnect_delay < MQTT_MAX_RECONNECT_DELAY
                ):  # first x retries are not exponential
                    self._reconnect_delay *= 2
                    self._reconnect_delay += uniform(0, 1)  # noqa: S311
                    self._reconnect_delay = min(self._reconnect_delay, MQTT_MAX_RECONNECT_DELAY)
                    _LOGGER.debug("Increased reconnect backoff to %s", self._reconnect_delay)

    def _on_message(self, msg: aiomqtt.Message) -> None:
        """Deserialize received MQTT message and emit Event to subscribed callbacks."""
        # Cast the data from binary string, ignoring empty messages.
        payload = cast("str", msg.payload)
        if len(payload) == 0:
            return

        topic = str(msg.topic)

        _LOGGER.debug("Message received on topic %s: %s", topic, payload)

        try:
            self._emit(BaseEvent.from_mqtt_message(topic=topic, payload=payload))
        except Exception as exc:  # noqa: BLE001
            _LOGGER.warning("Exception parsing MQTT event: %s", exc)

    def _emit(self, event: BaseEvent) -> None:
        for callback in self._callbacks:
            result = callback(event)
            if result is not None:
                task = asyncio.create_task(result)
                background_tasks.add(task)
                task.add_done_callback(background_tasks.discard)

        if isinstance(event, OperationEvent):
            self._handle_operation(event)

    def _handle_operation(self, event: OperationEvent) -> None:
        if event.status == OperationStatus.IN_PROGRESS:
            _LOGGER.debug(
                "An operation '%s' is now in progress. Trace id: %s",
                event.operation,
                event.trace_id,
            )
            return

        _LOGGER.debug(
            "Operation '%s' for trace id '%s' completed.",
            event.operation,
            event.trace_id,
        )
        self._handle_operation_completed(event)

    def _handle_operation_completed(self, event: OperationEvent) -> None:
        listeners = self._operation_listeners
        self._operation_listeners = []
        for listener in listeners:
            if listener.operation_name != event.operation:
                self._operation_listeners.append(listener)
                continue

            if event.status == OperationStatus.ERROR:
                _LOGGER.error(
                    "Resolving listener for operation '%s' with error '%s'.",
                    event.operation,
                    event.error_code,
                )
                listener.future.set_exception(OperationFailedError(event))
            else:
                if event.status == OperationStatus.COMPLETED_WARNING:
                    _LOGGER.warning("Operation '%s' completed with warnings.", event.operation)

                _LOGGER.debug("Resolving listener for operation '%s'.", event.operation)
                listener.future.set_result(event)
